@@ -1,9 +1,41 @@
+/*
+ * Copyright (c) 1990 - 1994, Julianne Frances Haugh
+ * Copyright (c) 1996 - 2001, Marek Michałkiewicz
+ * Copyright (c) 2001 - 2006, Tomasz Kłoczko
+ * Copyright (c) 2007 - 2011, Nicolas François
+ * All rights reserved.
+ *
+ * Redistribution and use in source and binary forms, with or without
+ * modification, are permitted provided that the following conditions
+ * are met:
+ * 1. Redistributions of source code must retain the above copyright
+ *    notice, this list of conditions and the following disclaimer.
+ * 2. Redistributions in binary form must reproduce the above copyright
+ *    notice, this list of conditions and the following disclaimer in the
+ *    documentation and/or other materials provided with the distribution.
+ * 3. The name of the copyright holders or contributors may not be used to
+ *    endorse or promote products derived from this software without
+ *    specific prior written permission.
+ *
+ * THIS SOFTWARE IS PROVIDED BY THE COPYRIGHT HOLDERS AND CONTRIBUTORS
+ * ``AS IS'' AND ANY EXPRESS OR IMPLIED WARRANTIES, INCLUDING, BUT NOT
+ * LIMITED TO, THE IMPLIED WARRANTIES OF MERCHANTABILITY AND FITNESS FOR A
+ * PARTICULAR PURPOSE ARE DISCLAIMED.  IN NO EVENT SHALL THE COPYRIGHT
+ * HOLDERS OR CONTRIBUTORS BE LIABLE FOR ANY DIRECT, INDIRECT, INCIDENTAL,
+ * SPECIAL, EXEMPLARY, OR CONSEQUENTIAL DAMAGES (INCLUDING, BUT NOT
+ * LIMITED TO, PROCUREMENT OF SUBSTITUTE GOODS OR SERVICES; LOSS OF USE,
+ * DATA, OR PROFITS; OR BUSINESS INTERRUPTION) HOWEVER CAUSED AND ON ANY
+ * THEORY OF LIABILITY, WHETHER IN CONTRACT, STRICT LIABILITY, OR TORT
+ * (INCLUDING NEGLIGENCE OR OTHERWISE) ARISING IN ANY WAY OUT OF THE USE
+ * OF THIS SOFTWARE, EVEN IF ADVISED OF THE POSSIBILITY OF SUCH DAMAGE.
+ */
 
 #include <config.h>
 
-#ident "$Id: commonio.c 1892 2008-03-08 22:52:44Z nekral-guest $"
+#ident "$Id: commonio.c 3727 2012-05-18 19:44:53Z nekral-guest $"
 
 #include "defines.h"
+#include <assert.h>
 #include <sys/stat.h>
 #include <stdlib.h>
 #include <limits.h>
@@ -13,30 +45,37 @@
 #include <stdio.h>
 #include <signal.h>
 #include "nscd.h"
-#ifdef WITH_SELINUX
-#include <selinux/selinux.h>
-static security_context_t old_context = NULL;
-#endif
+#ifdef WITH_TCB
+#include <tcb.h>
+#endif				/* WITH_TCB */
+#include "prototypes.h"
 #include "commonio.h"
 
 /* local function prototypes */
 static int lrename (const char *, const char *);
-static int check_link_count (const char *);
-static int do_lock_file (const char *, const char *);
-static FILE *fopen_set_perms (const char *, const char *, const struct stat *);
+static int check_link_count (const char *file);
+static int do_lock_file (const char *file, const char *lock, bool log);
+static /*@null@*/ /*@dependent@*/FILE *fopen_set_perms (
+	const char *name,
+	const char *mode,
+	const struct stat *sb);
 static int create_backup (const char *, FILE *);
 static void free_linked_list (struct commonio_db *);
-static void add_one_entry (struct commonio_db *, struct commonio_entry *);
-static int name_is_nis (const char *);
+static void add_one_entry (
+	struct commonio_db *db,
+	/*@owned@*/struct commonio_entry *p);
+static bool name_is_nis (const char *name);
 static int write_all (const struct commonio_db *);
-static struct commonio_entry *find_entry_by_name (struct commonio_db *,
-						  const char *);
-static struct commonio_entry *next_entry_by_name (struct commonio_db *,
-                                                  struct commonio_entry *pos,
-                                                  const char *);
+static /*@dependent@*/ /*@null@*/struct commonio_entry *find_entry_by_name (
+	struct commonio_db *,
+	const char *);
+static /*@dependent@*/ /*@null@*/struct commonio_entry *next_entry_by_name (
+	struct commonio_db *,
+	/*@null@*/struct commonio_entry *pos,
+	const char *);
 
 static int lock_count = 0;
-static int nscd_need_reload = 0;
+static bool nscd_need_reload = false;
 
 /*
  * Simple rename(P) alternative that attempts to rename to symlink
@@ -44,21 +83,36 @@ static int nscd_need_reload = 0;
  */
 int lrename (const char *old, const char *new)
 {
-
-	char resolved_path[PATH_MAX];
 	int res;
+	char *r = NULL;
 
 #if defined(S_ISLNK)
+#ifndef __GLIBC__
+	char resolved_path[PATH_MAX];
+#endif				/* !__GLIBC__ */
 	struct stat sb;
 	if (lstat (new, &sb) == 0 && S_ISLNK (sb.st_mode)) {
-		if (realpath (new, resolved_path) == NULL) {
+#ifdef __GLIBC__ /* now a POSIX.1-2008 feature */
+		r = realpath (new, NULL);
+#else				/* !__GLIBC__ */
+		r = realpath (new, resolved_path);
+#endif				/* !__GLIBC__ */
+		if (NULL == r) {
 			perror ("realpath in lrename()");
 		} else {
-			new = resolved_path;
+			new = r;
 		}
 	}
-#endif
+#endif				/* S_ISLNK */
+
 	res = rename (old, new);
+
+#ifdef __GLIBC__
+	if (NULL != r) {
+		free (r);
+	}
+#endif				/* __GLIBC__ */
+
 	return res;
 }
 
@@ -66,32 +120,46 @@ static int check_link_count (const char *file)
 {
 	struct stat sb;
 
-	if (stat (file, &sb) != 0)
+	if (stat (file, &sb) != 0) {
 		return 0;
+	}
 
-	if (sb.st_nlink != 2)
+	if (sb.st_nlink != 2) {
 		return 0;
+	}
 
 	return 1;
 }
 
 
-static int do_lock_file (const char *file, const char *lock)
+static int do_lock_file (const char *file, const char *lock, bool log)
 {
 	int fd;
-	int pid;
-	int len;
+	pid_t pid;
+	ssize_t len;
 	int retval;
 	char buf[32];
 
-	if ((fd = open (file, O_CREAT | O_EXCL | O_WRONLY, 0600)) == -1)
+	fd = open (file, O_CREAT | O_EXCL | O_WRONLY, 0600);
+	if (-1 == fd) {
+		if (log) {
+			(void) fprintf (stderr,
+			                "%s: %s: %s\n",
+			                Prog, file, strerror (errno));
+		}
 		return 0;
+	}
 
 	pid = getpid ();
-	snprintf (buf, sizeof buf, "%d", pid);
-	len = strlen (buf) + 1;
-	if (write (fd, buf, len) != len) {
-		close (fd);
+	snprintf (buf, sizeof buf, "%lu", (unsigned long) pid);
+	len = (ssize_t) strlen (buf) + 1;
+	if (write (fd, buf, (size_t) len) != len) {
+		if (log) {
+			(void) fprintf (stderr,
+			                "%s: %s: %s\n",
+			                Prog, file, strerror (errno));
+		}
+		(void) close (fd);
 		unlink (file);
 		return 0;
 	}
@@ -99,11 +167,22 @@ static int do_lock_file (const char *file, const char *lock)
 
 	if (link (file, lock) == 0) {
 		retval = check_link_count (file);
+		if ((0==retval) && log) {
+			(void) fprintf (stderr,
+			                "%s: %s: lock file already used\n",
+			                Prog, file);
+		}
 		unlink (file);
 		return retval;
 	}
 
-	if ((fd = open (lock, O_RDWR)) == -1) {
+	fd = open (lock, O_RDWR);
+	if (-1 == fd) {
+		if (log) {
+			(void) fprintf (stderr,
+			                "%s: %s: %s\n",
+			                Prog, lock, strerror (errno));
+		}
 		unlink (file);
 		errno = EINVAL;
 		return 0;
@@ -111,67 +190,107 @@ static int do_lock_file (const char *file, const char *lock)
 	len = read (fd, buf, sizeof (buf) - 1);
 	close (fd);
 	if (len <= 0) {
+		if (log) {
+			(void) fprintf (stderr,
+			                "%s: existing lock file %s without a PID\n",
+			                Prog, lock);
+		}
 		unlink (file);
 		errno = EINVAL;
 		return 0;
 	}
 	buf[len] = '\0';
-	if ((pid = strtol (buf, (char **) 0, 10)) == 0) {
+	if (get_pid (buf, &pid) == 0) {
+		if (log) {
+			(void) fprintf (stderr,
+			                "%s: existing lock file %s with an invalid PID '%s'\n",
+			                Prog, lock, buf);
+		}
 		unlink (file);
 		errno = EINVAL;
 		return 0;
 	}
 	if (kill (pid, 0) == 0) {
+		if (log) {
+			(void) fprintf (stderr,
+			                "%s: lock %s already used by PID %lu\n",
+			                Prog, lock, (unsigned long) pid);
+		}
 		unlink (file);
 		errno = EEXIST;
 		return 0;
 	}
 	if (unlink (lock) != 0) {
+		if (log) {
+			(void) fprintf (stderr,
+			                "%s: cannot get lock %s: %s\n",
+			                Prog, lock, strerror (errno));
+		}
 		unlink (file);
 		return 0;
 	}
 
 	retval = 0;
-	if (link (file, lock) == 0 && check_link_count (file))
-		retval = 1;
+	if (link (file, lock) == 0) {
+		retval = check_link_count (file);
+		if ((0==retval) && log) {
+			(void) fprintf (stderr,
+			                "%s: %s: lock file already used\n",
+			                Prog, file);
+		}
+	} else {
+		if (log) {
+			(void) fprintf (stderr,
+			                "%s: cannot get lock %s: %s\n",
+			                Prog, lock, strerror (errno));
+		}
+	}
 
 	unlink (file);
 	return retval;
 }
 
 
-static FILE *fopen_set_perms (const char *name, const char *mode,
-			      const struct stat *sb)
+static /*@null@*/ /*@dependent@*/FILE *fopen_set_perms (
+	const char *name,
+	const char *mode,
+	const struct stat *sb)
 {
 	FILE *fp;
 	mode_t mask;
 
 	mask = umask (0777);
 	fp = fopen (name, mode);
-	umask (mask);
-	if (!fp)
+	(void) umask (mask);
+	if (NULL == fp) {
 		return NULL;
+	}
 
 #ifdef HAVE_FCHOWN
-	if (fchown (fileno (fp), sb->st_uid, sb->st_gid))
+	if (fchown (fileno (fp), sb->st_uid, sb->st_gid) != 0) {
 		goto fail;
-#else
-	if (chown (name, sb->st_mode))
+	}
+#else				/* !HAVE_FCHOWN */
+	if (chown (name, sb->st_mode) != 0) {
 		goto fail;
-#endif
+	}
+#endif				/* !HAVE_FCHOWN */
 
 #ifdef HAVE_FCHMOD
-	if (fchmod (fileno (fp), sb->st_mode & 0664))
+	if (fchmod (fileno (fp), sb->st_mode & 0664) != 0) {
 		goto fail;
-#else
-	if (chmod (name, sb->st_mode & 0664))
+	}
+#else				/* !HAVE_FCHMOD */
+	if (chmod (name, sb->st_mode & 0664) != 0) {
 		goto fail;
-#endif
+	}
+#endif				/* !HAVE_FCHMOD */
 	return fp;
 
       fail:
-	fclose (fp);
-	unlink (name);
+	(void) fclose (fp);
+	/* fopen_set_perms is used for intermediate files */
+	(void) unlink (name);
 	return NULL;
 }
 
@@ -184,32 +303,40 @@ static int create_backup (const char *backup, FILE * fp)
 	int c;
 	mode_t mask;
 
-	if (fstat (fileno (fp), &sb))
+	if (fstat (fileno (fp), &sb) != 0) {
 		return -1;
+	}
 
 	mask = umask (077);
 	bkfp = fopen (backup, "w");
-	umask (mask);
-	if (!bkfp)
+	(void) umask (mask);
+	if (NULL == bkfp) {
 		return -1;
+	}
 
 	/* TODO: faster copy, not one-char-at-a-time.  --marekm */
 	c = 0;
-	if (fseek (fp, 0, SEEK_SET) == 0)
+	if (fseek (fp, 0, SEEK_SET) == 0) {
 		while ((c = getc (fp)) != EOF) {
-			if (putc (c, bkfp) == EOF)
+			if (putc (c, bkfp) == EOF) {
 				break;
+			}
 		}
-	if (c != EOF || ferror (fp) || fflush (bkfp)) {
-		fclose (bkfp);
+	}
+	if ((c != EOF) || (ferror (fp) != 0) || (fflush (bkfp) != 0)) {
+		(void) fclose (bkfp);
+		/* FIXME: unlink the backup file? */
 		return -1;
 	}
-	if (fclose (bkfp))
+	if (   (fsync (fileno (bkfp)) != 0)
+	    || (fclose (bkfp) != 0)) {
+		/* FIXME: unlink the backup file? */
 		return -1;
+	}
 
 	ub.actime = sb.st_atime;
 	ub.modtime = sb.st_mtime;
-	utime (backup, &ub);
+	(void) utime (backup, &ub);
 	return 0;
 }
 
@@ -218,15 +345,17 @@ static void free_linked_list (struct commonio_db *db)
 {
 	struct commonio_entry *p;
 
-	while (db->head) {
+	while (NULL != db->head) {
 		p = db->head;
 		db->head = p->next;
 
-		if (p->line)
+		if (NULL != p->line) {
 			free (p->line);
+		}
 
-		if (p->eptr)
+		if (NULL != p->eptr) {
 			db->ops->free (p->eptr);
+		}
 
 		free (p);
 	}
@@ -241,24 +370,26 @@ int commonio_setname (struct commonio_db *db, const char *name)
 }
 
 
-int commonio_present (const struct commonio_db *db)
+bool commonio_present (const struct commonio_db *db)
 {
 	return (access (db->filename, F_OK) == 0);
 }
 
 
-int commonio_lock_nowait (struct commonio_db *db)
+int commonio_lock_nowait (struct commonio_db *db, bool log)
 {
 	char file[1024];
 	char lock[1024];
 
-	if (db->locked)
+	if (db->locked) {
 		return 1;
+	}
 
-	snprintf (file, sizeof file, "%s.%ld", db->filename, (long) getpid ());
+	snprintf (file, sizeof file, "%s.%lu",
+	          db->filename, (unsigned long) getpid ());
 	snprintf (lock, sizeof lock, "%s.lock", db->filename);
-	if (do_lock_file (file, lock)) {
-		db->locked = 1;
+	if (do_lock_file (file, lock, log) != 0) {
+		db->locked = true;
 		lock_count++;
 		return 1;
 	}
@@ -279,17 +410,24 @@ int commonio_lock (struct commonio_db *db)
 	 * If it succeeds, call *_lock() only once
 	 * (no retries, it should always succeed).
 	 */
-	if (lock_count == 0) {
-		if (lckpwdf () == -1)
+	if (0 == lock_count) {
+		if (lckpwdf () == -1) {
+			if (geteuid () != 0) {
+				(void) fprintf (stderr,
+				                "%s: Permission denied.\n",
+				                Prog);
+			}
 			return 0;	/* failure */
+		}
 	}
 
-	if (commonio_lock_nowait (db))
+	if (commonio_lock_nowait (db, true) != 0) {
 		return 1;	/* success */
+	}
 
 	ulckpwdf ();
 	return 0;		/* failure */
-#else
+#else				/* !HAVE_LCKPWDF */
 	int i;
 
 	/*
@@ -303,16 +441,21 @@ int commonio_lock (struct commonio_db *db)
 #define LOCK_SLEEP 1
 #endif
 	for (i = 0; i < LOCK_TRIES; i++) {
-		if (i > 0)
+		if (i > 0) {
 			sleep (LOCK_SLEEP);	/* delay between retries */
-		if (commonio_lock_nowait (db))
+		}
+		if (commonio_lock_nowait (db, i==LOCK_TRIES-1) != 0) {
 			return 1;	/* success */
+		}
 		/* no unnecessary retries on "permission denied" errors */
-		if (geteuid () != 0)
+		if (geteuid () != 0) {
+			(void) fprintf (stderr, "%s: Permission denied.\n",
+			                Prog);
 			return 0;
+		}
 	}
 	return 0;		/* failure */
-#endif
+#endif				/* !HAVE_LCKPWDF */
 }
 
 static void dec_lock_count (void)
@@ -325,11 +468,11 @@ static void dec_lock_count (void)
 			if (nscd_need_reload) {
 				nscd_flush_cache ("passwd");
 				nscd_flush_cache ("group");
-				nscd_need_reload = 0;
+				nscd_need_reload = false;
 			}
 #ifdef HAVE_LCKPWDF
 			ulckpwdf ();
-#endif
+#endif				/* HAVE_LCKPWDF */
 		}
 	}
 }
@@ -340,10 +483,11 @@ int commonio_unlock (struct commonio_db *db)
 	char lock[1024];
 
 	if (db->isopen) {
-		db->readonly = 1;
-		if (!commonio_close (db)) {
-			if (db->locked)
+		db->readonly = true;
+		if (commonio_close (db) == 0) {
+			if (db->locked) {
 				dec_lock_count ();
+			}
 			return 0;
 		}
 	}
@@ -352,7 +496,7 @@ int commonio_unlock (struct commonio_db *db)
 		 * Unlock in reverse order: remove the lock file,
 		 * then call ulckpwdf() (if used) on last unlock.
 		 */
-		db->locked = 0;
+		db->locked = false;
 		snprintf (lock, sizeof lock, "%s.lock", db->filename);
 		unlink (lock);
 		dec_lock_count ();
@@ -362,21 +506,32 @@ int commonio_unlock (struct commonio_db *db)
 }
 
 
-static void add_one_entry (struct commonio_db *db, struct commonio_entry *p)
+/*
+ * Add an entry at the end.
+ *
+ * defines p->next, p->prev
+ * (unfortunately, owned special are not supported)
+ */
+static void add_one_entry (struct commonio_db *db,
+                           /*@owned@*/struct commonio_entry *p)
 {
+	/*@-mustfreeonly@*/
 	p->next = NULL;
 	p->prev = db->tail;
-	if (!db->head)
+	/*@=mustfreeonly@*/
+	if (NULL == db->head) {
 		db->head = p;
-	if (db->tail)
+	}
+	if (NULL != db->tail) {
 		db->tail->next = p;
+	}
 	db->tail = p;
 }
 
 
-static int name_is_nis (const char *n)
+static bool name_is_nis (const char *name)
 {
-	return (n[0] == '+' || n[0] == '-');
+	return (('+' == name[0]) || ('-' == name[0]));
 }
 
 
@@ -389,25 +544,32 @@ static int name_is_nis (const char *n)
 #endif
 
 #if KEEP_NIS_AT_END
-static void add_one_entry_nis (struct commonio_db *, struct commonio_entry *);
+static void add_one_entry_nis (struct commonio_db *db,
+                               /*@owned@*/struct commonio_entry *newp);
 
 /*
  * Insert an entry between the regular entries, and the NIS entries.
+ *
+ * defines newp->next, newp->prev
+ * (unfortunately, owned special are not supported)
  */
-static void
-add_one_entry_nis (struct commonio_db *db, struct commonio_entry *newp)
+static void add_one_entry_nis (struct commonio_db *db,
+                               /*@owned@*/struct commonio_entry *newp)
 {
 	struct commonio_entry *p;
 
-	for (p = db->head; p; p = p->next) {
-		if (name_is_nis
-		    (p->eptr ? db->ops->getname (p->eptr) : p->line)) {
+	for (p = db->head; NULL != p; p = p->next) {
+		if (name_is_nis (p->eptr ? db->ops->getname (p->eptr)
+		                         : p->line)) {
+			/*@-mustfreeonly@*/
 			newp->next = p;
 			newp->prev = p->prev;
-			if (p->prev)
+			/*@=mustfreeonly@*/
+			if (NULL != p->prev) {
 				p->prev->next = newp;
-			else
+			} else {
 				db->head = newp;
+			}
 			p->prev = newp;
 			return;
 		}
@@ -426,14 +588,17 @@ int commonio_open (struct commonio_db *db, int mode)
 	char *cp;
 	char *line;
 	struct commonio_entry *p;
-	void *eptr;
+	void *eptr = NULL;
 	int flags = mode;
-	int buflen;
+	size_t buflen;
+	int fd;
 	int saved_errno;
 
 	mode &= ~O_CREAT;
 
-	if (db->isopen || (mode != O_RDONLY && mode != O_RDWR)) {
+	if (   db->isopen
+	    || (   (O_RDONLY != mode)
+	        && (O_RDWR != mode))) {
 		errno = EINVAL;
 		return 0;
 	}
@@ -443,91 +608,122 @@ int commonio_open (struct commonio_db *db, int mode)
 		return 0;
 	}
 
-	db->head = db->tail = db->cursor = NULL;
-	db->changed = 0;
+	db->head = NULL;
+	db->tail = NULL;
+	db->cursor = NULL;
+	db->changed = false;
 
-	db->fp = fopen (db->filename, db->readonly ? "r" : "r+");
+	fd = open (db->filename,
+	             (db->readonly ? O_RDONLY : O_RDWR)
+	           | O_NOCTTY | O_NONBLOCK | O_NOFOLLOW);
+	saved_errno = errno;
+	db->fp = NULL;
+	if (fd >= 0) {
+#ifdef WITH_TCB
+		if (tcb_is_suspect (fd) != 0) {
+			(void) close (fd);
+			errno = EINVAL;
+			return 0;
+		}
+#endif				/* WITH_TCB */
+		db->fp = fdopen (fd, db->readonly ? "r" : "r+");
+		saved_errno = errno;
+		if (NULL == db->fp) {
+			(void) close (fd);
+		}
+	}
+	errno = saved_errno;
 
 	/*
 	 * If O_CREAT was specified and the file didn't exist, it will be
 	 * created by commonio_close().  We have no entries to read yet.  --marekm
 	 */
-	if (!db->fp) {
-		if ((flags & O_CREAT) && errno == ENOENT) {
-			db->isopen = 1;
+	if (NULL == db->fp) {
+		if (((flags & O_CREAT) != 0) && (ENOENT == errno)) {
+			db->isopen = true;
 			return 1;
 		}
 		return 0;
 	}
 
 	/* Do not inherit fd in spawned processes (e.g. nscd) */
-	fcntl(fileno(db->fp), F_SETFD, FD_CLOEXEC);
-
-#ifdef WITH_SELINUX
-	db->scontext = NULL;
-	if ((is_selinux_enabled () > 0) && (!db->readonly)) {
-		if (fgetfilecon (fileno (db->fp), &db->scontext) < 0) {
-			goto cleanup_errno;
-		}
-	}
-#endif
+	fcntl (fileno (db->fp), F_SETFD, FD_CLOEXEC);
 
 	buflen = BUFLEN;
 	buf = (char *) malloc (buflen);
-	if (!buf)
+	if (NULL == buf) {
 		goto cleanup_ENOMEM;
+	}
 
-	while (db->ops->fgets (buf, buflen, db->fp)) {
-		while (!(cp = strrchr (buf, '\n')) && !feof (db->fp)) {
-			int len;
+	while (db->ops->fgets (buf, (int) buflen, db->fp) == buf) {
+		while (   ((cp = strrchr (buf, '\n')) == NULL)
+		       && (feof (db->fp) == 0)) {
+			size_t len;
 
 			buflen += BUFLEN;
 			cp = (char *) realloc (buf, buflen);
-			if (!cp)
+			if (NULL == cp) {
 				goto cleanup_buf;
+			}
 			buf = cp;
 			len = strlen (buf);
-			db->ops->fgets (buf + len, buflen - len, db->fp);
+			if (db->ops->fgets (buf + len,
+			                    (int) (buflen - len),
+			                    db->fp) == NULL) {
+				goto cleanup_buf;
+			}
 		}
-		if ((cp = strrchr (buf, '\n')))
+		cp = strrchr (buf, '\n');
+		if (NULL != cp) {
 			*cp = '\0';
+		}
 
-		if (!(line = strdup (buf)))
+		line = strdup (buf);
+		if (NULL == line) {
 			goto cleanup_buf;
+		}
 
 		if (name_is_nis (line)) {
 			eptr = NULL;
-		} else if ((eptr = db->ops->parse (line))) {
-			eptr = db->ops->dup (eptr);
-			if (!eptr)
-				goto cleanup_line;
+		} else {
+			eptr = db->ops->parse (line);
+			if (NULL != eptr) {
+				eptr = db->ops->dup (eptr);
+				if (NULL == eptr) {
+					goto cleanup_line;
+				}
+			}
 		}
 
 		p = (struct commonio_entry *) malloc (sizeof *p);
-		if (!p)
+		if (NULL == p) {
 			goto cleanup_entry;
+		}
 
 		p->eptr = eptr;
 		p->line = line;
-		p->changed = 0;
+		p->changed = false;
 
 		add_one_entry (db, p);
 	}
 
 	free (buf);
 
-	if (ferror (db->fp))
+	if (ferror (db->fp) != 0) {
 		goto cleanup_errno;
+	}
 
-	if (db->ops->open_hook && !db->ops->open_hook ())
+	if ((NULL != db->ops->open_hook) && (db->ops->open_hook () == 0)) {
 		goto cleanup_errno;
+	}
 
-	db->isopen = 1;
+	db->isopen = true;
 	return 1;
 
       cleanup_entry:
-	if (eptr)
+	if (NULL != eptr) {
 		db->ops->free (eptr);
+	}
       cleanup_line:
 	free (line);
       cleanup_buf:
@@ -537,12 +733,6 @@ int commonio_open (struct commonio_db *db, int mode)
       cleanup_errno:
 	saved_errno = errno;
 	free_linked_list (db);
-#ifdef WITH_SELINUX
-	if (db->scontext != NULL) {
-		freecon (db->scontext);
-		db->scontext = NULL;
-	}
-#endif
 	fclose (db->fp);
 	db->fp = NULL;
 	errno = saved_errno;
@@ -556,37 +746,79 @@ int
 commonio_sort (struct commonio_db *db, int (*cmp) (const void *, const void *))
 {
 	struct commonio_entry **entries, *ptr;
-	int n = 0, i;
+	size_t n = 0, i;
+#if KEEP_NIS_AT_END
+	struct commonio_entry *nis = NULL;
+#endif
 
-	for (ptr = db->head; ptr; ptr = ptr->next)
+	for (ptr = db->head;
+	        (NULL != ptr)
+#if KEEP_NIS_AT_END
+	     && (NULL != ptr->line)
+	     && (   ('+' != ptr->line[0])
+	         && ('-' != ptr->line[0]))
+#endif
+	     ;
+	     ptr = ptr->next) {
 		n++;
+	}
+#if KEEP_NIS_AT_END
+	if ((NULL != ptr) && (NULL != ptr->line)) {
+		nis = ptr;
+	}
+#endif
 
-	if (n <= 1)
+	if (n <= 1) {
 		return 0;
+	}
 
 	entries = malloc (n * sizeof (struct commonio_entry *));
-	if (entries == NULL)
+	if (entries == NULL) {
 		return -1;
+	}
 
 	n = 0;
-	for (ptr = db->head; ptr; ptr = ptr->next)
-		entries[n++] = ptr;
+	for (ptr = db->head;
+#if KEEP_NIS_AT_END
+	     nis != ptr;
+#else
+	     NULL != ptr;
+#endif
+/*@ -nullderef @*/
+	     ptr = ptr->next
+/*@ +nullderef @*/
+	    ) {
+		entries[n] = ptr;
+		n++;
+	}
 	qsort (entries, n, sizeof (struct commonio_entry *), cmp);
 
+	/* Take care of the head and tail separately */
 	db->head = entries[0];
-	db->tail = entries[--n];
+	n--;
+#if KEEP_NIS_AT_END
+	if (NULL == nis)
+#endif
+	{
+		db->tail = entries[n];
+	}
 	db->head->prev = NULL;
 	db->head->next = entries[1];
-	db->tail->prev = entries[n - 1];
-	db->tail->next = NULL;
+	entries[n]->prev = entries[n - 1];
+#if KEEP_NIS_AT_END
+	entries[n]->next = nis;
+#else
+	entries[n]->next = NULL;
+#endif
 
+	/* Now other elements have prev and next entries */
 	for (i = 1; i < n; i++) {
 		entries[i]->prev = entries[i - 1];
 		entries[i]->next = entries[i + 1];
 	}
 
 	free (entries);
-	db->changed = 1;
+	db->changed = true;
 
 	return 0;
 }
@@ -594,40 +826,52 @@ commonio_sort (struct commonio_db *db, int (*cmp) (const void *, const void *))
 /*
  * Sort entries in db according to order in another.
  */
-int commonio_sort_wrt (struct commonio_db *shadow, struct commonio_db *passwd)
+int commonio_sort_wrt (struct commonio_db *shadow,
+                       const struct commonio_db *passwd)
 {
 	struct commonio_entry *head = NULL, *pw_ptr, *spw_ptr;
 	const char *name;
 
-	if (!shadow || !shadow->head)
+	if ((NULL == shadow) || (NULL == shadow->head)) {
 		return 0;
+	}
 
-	for (pw_ptr = passwd->head; pw_ptr; pw_ptr = pw_ptr->next) {
-		if (pw_ptr->eptr == NULL)
+	for (pw_ptr = passwd->head; NULL != pw_ptr; pw_ptr = pw_ptr->next) {
+		if (NULL == pw_ptr->eptr) {
 			continue;
+		}
 		name = passwd->ops->getname (pw_ptr->eptr);
-		for (spw_ptr = shadow->head; spw_ptr; spw_ptr = spw_ptr->next)
+		for (spw_ptr = shadow->head;
+		     NULL != spw_ptr;
+		     spw_ptr = spw_ptr->next) {
+			if (NULL == spw_ptr->eptr) {
+				continue;
+			}
 			if (strcmp (name, shadow->ops->getname (spw_ptr->eptr))
-			    == 0)
+			    == 0) {
 				break;
-		if (spw_ptr == NULL)
+			}
+		}
+		if (NULL == spw_ptr) {
 			continue;
+		}
 		commonio_del_entry (shadow, spw_ptr);
 		spw_ptr->next = head;
 		head = spw_ptr;
 	}
 
-	for (spw_ptr = head; spw_ptr; spw_ptr = head) {
+	for (spw_ptr = head; NULL != spw_ptr; spw_ptr = head) {
 		head = head->next;
 
-		if (shadow->head)
+		if (NULL != shadow->head) {
 			shadow->head->prev = spw_ptr;
+		}
 		spw_ptr->next = shadow->head;
 		shadow->head = spw_ptr;
 	}
 
 	shadow->head->prev = NULL;
-	shadow->changed = 1;
+	shadow->changed = true;
 
 	return 0;
 }
@@ -635,23 +879,28 @@ int commonio_sort_wrt (struct commonio_db *shadow, struct commonio_db *passwd)
 /*
  * write_all - Write the database to its file.
  *
- * It returns 0 if all the entries could be writen correctly.
+ * It returns 0 if all the entries could be written correctly.
  */
 static int write_all (const struct commonio_db *db)
+	/*@requires notnull db->fp@*/
 {
 	const struct commonio_entry *p;
 	void *eptr;
 
-	for (p = db->head; p; p = p->next) {
+	for (p = db->head; NULL != p; p = p->next) {
 		if (p->changed) {
 			eptr = p->eptr;
-			if (db->ops->put (eptr, db->fp))
+			assert (NULL != eptr);
+			if (db->ops->put (eptr, db->fp) != 0) {
 				return -1;
-		} else if (p->line) {
-			if (db->ops->fputs (p->line, db->fp) == EOF)
+			}
+		} else if (NULL != p->line) {
+			if (db->ops->fputs (p->line, db->fp) == EOF) {
 				return -1;
-			if (putc ('\n', db->fp) == EOF)
+			}
+			if (putc ('\n', db->fp) == EOF) {
 				return -1;
+			}
 		}
 	}
 	return 0;
@@ -659,6 +908,7 @@ static int write_all (const struct commonio_db *db)
 
 
 int commonio_close (struct commonio_db *db)
+	/*@requires notnull db->fp@*/
 {
 	char buf[1024];
 	int errors = 0;
@@ -668,48 +918,50 @@ int commonio_close (struct commonio_db *db)
 		errno = EINVAL;
 		return 0;
 	}
-	db->isopen = 0;
+	db->isopen = false;
 
 	if (!db->changed || db->readonly) {
-		fclose (db->fp);
+		(void) fclose (db->fp);
 		db->fp = NULL;
 		goto success;
 	}
 
-	if (db->ops->close_hook && !db->ops->close_hook ())
+	if ((NULL != db->ops->close_hook) && (db->ops->close_hook () == 0)) {
 		goto fail;
+	}
 
 	memzero (&sb, sizeof sb);
-	if (db->fp) {
-		if (fstat (fileno (db->fp), &sb)) {
-			fclose (db->fp);
+	if (NULL != db->fp) {
+		if (fstat (fileno (db->fp), &sb) != 0) {
+			(void) fclose (db->fp);
 			db->fp = NULL;
 			goto fail;
 		}
-#ifdef WITH_SELINUX
-		if (db->scontext != NULL) {
-			if (getfscreatecon (&old_context) < 0) {
-				errors++;
-				goto fail;
-			}
-			if (setfscreatecon (db->scontext) < 0) {
-				errors++;
-				goto fail;
-			}
-		}
-#endif
+
 		/*
 		 * Create backup file.
 		 */
 		snprintf (buf, sizeof buf, "%s-", db->filename);
 
-		if (create_backup (buf, db->fp))
+#ifdef WITH_SELINUX
+		if (set_selinux_file_context (buf) != 0) {
 			errors++;
-
-		if (fclose (db->fp))
+		}
+#endif
+		if (create_backup (buf, db->fp) != 0) {
 			errors++;
+		}
 
-		if (errors) {
+		if (fclose (db->fp) != 0) {
+			errors++;
+		}
+
+#ifdef WITH_SELINUX
+		if (reset_selinux_file_context () != 0) {
+			errors++;
+		}
+#endif
+		if (errors != 0) {
 			db->fp = NULL;
 			goto fail;
 		}
@@ -725,79 +977,89 @@ int commonio_close (struct commonio_db *db)
 
 	snprintf (buf, sizeof buf, "%s+", db->filename);
 
-	db->fp = fopen_set_perms (buf, "w", &sb);
-	if (!db->fp)
-		goto fail;
-
-	if (write_all (db))
+#ifdef WITH_SELINUX
+	if (set_selinux_file_context (buf) != 0) {
 		errors++;
-
-	if (fflush (db->fp))
-		errors++;
-#ifdef HAVE_FSYNC
-	if (fsync (fileno (db->fp)))
-		errors++;
-#else
-	sync ();
+	}
 #endif
-	if (fclose (db->fp))
+
+	db->fp = fopen_set_perms (buf, "w", &sb);
+	if (NULL == db->fp) {
+		goto fail;
+	}
+
+	if (write_all (db) != 0) {
 		errors++;
+	}
+
+	if (fflush (db->fp) != 0) {
+		errors++;
+	}
+#ifdef HAVE_FSYNC
+	if (fsync (fileno (db->fp)) != 0) {
+		errors++;
+	}
+#else				/* !HAVE_FSYNC */
+	sync ();
+#endif				/* !HAVE_FSYNC */
+	if (fclose (db->fp) != 0) {
+		errors++;
+	}
 
 	db->fp = NULL;
 
-	if (errors) {
+	if (errors != 0) {
 		unlink (buf);
 		goto fail;
 	}
 
-	if (lrename (buf, db->filename))
+	if (lrename (buf, db->filename) != 0) {
 		goto fail;
+	}
 
-	nscd_need_reload = 1;
+#ifdef WITH_SELINUX
+	if (reset_selinux_file_context () != 0) {
+		goto fail;
+	}
+#endif
+
+	nscd_need_reload = true;
 	goto success;
       fail:
 	errors++;
       success:
 
-#ifdef WITH_SELINUX
-	if (db->scontext != NULL) {
-		if (setfscreatecon (old_context) < 0) {
-			errors++;
-		}
-		if (old_context != NULL) {
-			freecon (old_context);
-			old_context = NULL;
-		}
-		freecon (db->scontext);
-		db->scontext = NULL;
-	}
-#endif
 	free_linked_list (db);
 	return errors == 0;
 }
 
-static struct commonio_entry *next_entry_by_name (struct commonio_db *db,
-                                                  struct commonio_entry *pos,
-                                                  const char *name)
+static /*@dependent@*/ /*@null@*/struct commonio_entry *next_entry_by_name (
+	struct commonio_db *db,
+	/*@null@*/struct commonio_entry *pos,
+	const char *name)
 {
 	struct commonio_entry *p;
 	void *ep;
 
-	if (NULL == pos)
+	if (NULL == pos) {
 		return NULL;
+	}
 
-	for (p = pos; p; p = p->next) {
+	for (p = pos; NULL != p; p = p->next) {
 		ep = p->eptr;
-		if (ep && strcmp (db->ops->getname (ep), name) == 0)
+		if (   (NULL != ep)
+		    && (strcmp (db->ops->getname (ep), name) == 0)) {
 			break;
+		}
 	}
 	return p;
 }
 
-static struct commonio_entry *find_entry_by_name (struct commonio_db *db,
-						  const char *name)
+static /*@dependent@*/ /*@null@*/struct commonio_entry *find_entry_by_name (
+	struct commonio_db *db,
+	const char *name)
 {
-	return next_entry_by_name(db, db->head, name);
+	return next_entry_by_name (db, db->head, name);
 }
 
 
@@ -810,28 +1072,28 @@ int commonio_update (struct commonio_db *db, const void *eptr)
 		errno = EINVAL;
 		return 0;
 	}
-	if (!(nentry = db->ops->dup (eptr))) {
+	nentry = db->ops->dup (eptr);
+	if (NULL == nentry) {
 		errno = ENOMEM;
 		return 0;
 	}
 	p = find_entry_by_name (db, db->ops->getname (eptr));
-	if (p) {
-		if (next_entry_by_name (db, p->next, db->ops->getname (eptr)))
-		{
+	if (NULL != p) {
+		if (next_entry_by_name (db, p->next, db->ops->getname (eptr)) != NULL) {
 			fprintf (stderr, _("Multiple entries named '%s' in %s. Please fix this with pwck or grpck.\n"), db->ops->getname (eptr), db->filename);
 			return 0;
 		}
 		db->ops->free (p->eptr);
 		p->eptr = nentry;
-		p->changed = 1;
+		p->changed = true;
 		db->cursor = p;
 
-		db->changed = 1;
+		db->changed = true;
 		return 1;
 	}
 	/* not found, new entry */
 	p = (struct commonio_entry *) malloc (sizeof *p);
-	if (!p) {
+	if (NULL == p) {
 		db->ops->free (nentry);
 		errno = ENOMEM;
 		return 0;
@@ -839,35 +1101,38 @@ int commonio_update (struct commonio_db *db, const void *eptr)
 
 	p->eptr = nentry;
 	p->line = NULL;
-	p->changed = 1;
+	p->changed = true;
 
 #if KEEP_NIS_AT_END
 	add_one_entry_nis (db, p);
-#else
+#else				/* !KEEP_NIS_AT_END */
 	add_one_entry (db, p);
-#endif
+#endif				/* !KEEP_NIS_AT_END */
 
-	db->changed = 1;
+	db->changed = true;
 	return 1;
 }
 
 
 void commonio_del_entry (struct commonio_db *db, const struct commonio_entry *p)
 {
-	if (p == db->cursor)
+	if (p == db->cursor) {
 		db->cursor = p->next;
+	}
 
-	if (p->prev)
+	if (NULL != p->prev) {
 		p->prev->next = p->next;
-	else
+	} else {
 		db->head = p->next;
+	}
 
-	if (p->next)
+	if (NULL != p->next) {
 		p->next->prev = p->prev;
-	else
+	} else {
 		db->tail = p->prev;
+	}
 
-	db->changed = 1;
+	db->changed = true;
 }
 
 /*
@@ -882,22 +1147,24 @@ int commonio_remove (struct commonio_db *db, const char *name)
 		return 0;
 	}
 	p = find_entry_by_name (db, name);
-	if (!p) {
+	if (NULL == p) {
 		errno = ENOENT;
 		return 0;
 	}
-	if (next_entry_by_name (db, p->next, name)) {
+	if (next_entry_by_name (db, p->next, name) != NULL) {
 		fprintf (stderr, _("Multiple entries named '%s' in %s. Please fix this with pwck or grpck.\n"), name, db->filename);
 		return 0;
 	}
 
 	commonio_del_entry (db, p);
 
-	if (p->line)
+	if (NULL != p->line) {
 		free (p->line);
+	}
 
-	if (p->eptr)
+	if (NULL != p->eptr) {
 		db->ops->free (p->eptr);
+	}
 
 	return 1;
 }
@@ -911,7 +1178,7 @@ int commonio_remove (struct commonio_db *db, const char *name)
  *
  *	Otherwise, it returns NULL.
  */
-const void *commonio_locate (struct commonio_db *db, const char *name)
+/*@observer@*/ /*@null@*/const void *commonio_locate (struct commonio_db *db, const char *name)
 {
 	struct commonio_entry *p;
 
@@ -920,7 +1187,7 @@ const void *commonio_locate (struct commonio_db *db, const char *name)
 		return NULL;
 	}
 	p = find_entry_by_name (db, name);
-	if (!p) {
+	if (NULL == p) {
 		errno = ENOENT;
 		return NULL;
 	}
@@ -948,7 +1215,7 @@ int commonio_rewind (struct commonio_db *db)
  *
  * It returns the next entry, or NULL if no other entries could be found.
  */
-const void *commonio_next (struct commonio_db *db)
+/*@observer@*/ /*@null@*/const void *commonio_next (struct commonio_db *db)
 {
 	void *eptr;
 
@@ -956,17 +1223,20 @@ const void *commonio_next (struct commonio_db *db)
 		errno = EINVAL;
 		return 0;
 	}
-	if (db->cursor == NULL)
+	if (NULL == db->cursor) {
 		db->cursor = db->head;
-	else
+	} else {
 		db->cursor = db->cursor->next;
+	}
 
-	while (db->cursor) {
+	while (NULL != db->cursor) {
 		eptr = db->cursor->eptr;
-		if (eptr)
+		if (NULL != eptr) {
 			return eptr;
+		}
 
 		db->cursor = db->cursor->next;
 	}
 	return NULL;
 }
+
