@@ -1,5 +1,8 @@
 /*
- * Copyright 1991 - 1994, Julianne Frances Haugh and Chip Rosenthal
+ * Copyright (c) 1991 - 1994, Julianne Frances Haugh
+ * Copyright (c) 1996 - 2000, Marek Michałkiewicz
+ * Copyright (c) 2002 - 2006, Tomasz Kłoczko
+ * Copyright (c) 2007 - 2008, Nicolas François
  * All rights reserved.
  *
  * Redistribution and use in source and binary forms, with or without
@@ -10,39 +13,40 @@
  * 2. Redistributions in binary form must reproduce the above copyright
  *    notice, this list of conditions and the following disclaimer in the
  *    documentation and/or other materials provided with the distribution.
- * 3. Neither the name of Julianne F. Haugh nor the names of its contributors
- *    may be used to endorse or promote products derived from this software
- *    without specific prior written permission.
+ * 3. The name of the copyright holders or contributors may not be used to
+ *    endorse or promote products derived from this software without
+ *    specific prior written permission.
  *
- * THIS SOFTWARE IS PROVIDED BY JULIE HAUGH AND CONTRIBUTORS ``AS IS'' AND
- * ANY EXPRESS OR IMPLIED WARRANTIES, INCLUDING, BUT NOT LIMITED TO, THE
- * IMPLIED WARRANTIES OF MERCHANTABILITY AND FITNESS FOR A PARTICULAR PURPOSE
- * ARE DISCLAIMED.  IN NO EVENT SHALL JULIE HAUGH OR CONTRIBUTORS BE LIABLE
- * FOR ANY DIRECT, INDIRECT, INCIDENTAL, SPECIAL, EXEMPLARY, OR CONSEQUENTIAL
- * DAMAGES (INCLUDING, BUT NOT LIMITED TO, PROCUREMENT OF SUBSTITUTE GOODS
- * OR SERVICES; LOSS OF USE, DATA, OR PROFITS; OR BUSINESS INTERRUPTION)
- * HOWEVER CAUSED AND ON ANY THEORY OF LIABILITY, WHETHER IN CONTRACT, STRICT
- * LIABILITY, OR TORT (INCLUDING NEGLIGENCE OR OTHERWISE) ARISING IN ANY WAY
- * OUT OF THE USE OF THIS SOFTWARE, EVEN IF ADVISED OF THE POSSIBILITY OF
- * SUCH DAMAGE.
+ * THIS SOFTWARE IS PROVIDED BY THE COPYRIGHT HOLDERS AND CONTRIBUTORS
+ * ``AS IS'' AND ANY EXPRESS OR IMPLIED WARRANTIES, INCLUDING, BUT NOT
+ * LIMITED TO, THE IMPLIED WARRANTIES OF MERCHANTABILITY AND FITNESS FOR A
+ * PARTICULAR PURPOSE ARE DISCLAIMED.  IN NO EVENT SHALL THE COPYRIGHT
+ * HOLDERS OR CONTRIBUTORS BE LIABLE FOR ANY DIRECT, INDIRECT, INCIDENTAL,
+ * SPECIAL, EXEMPLARY, OR CONSEQUENTIAL DAMAGES (INCLUDING, BUT NOT
+ * LIMITED TO, PROCUREMENT OF SUBSTITUTE GOODS OR SERVICES; LOSS OF USE,
+ * DATA, OR PROFITS; OR BUSINESS INTERRUPTION) HOWEVER CAUSED AND ON ANY
+ * THEORY OF LIABILITY, WHETHER IN CONTRACT, STRICT LIABILITY, OR TORT
+ * (INCLUDING NEGLIGENCE OR OTHERWISE) ARISING IN ANY WAY OUT OF THE USE
+ * OF THIS SOFTWARE, EVEN IF ADVISED OF THE POSSIBILITY OF SUCH DAMAGE.
  */
 
 #include <config.h>
 
-#ident "$Id: getdef.c 1862 2008-02-25 21:06:30Z nekral-guest $"
+#ident "$Id: getdef.c 3095 2010-03-04 18:11:13Z nekral-guest $"
 
 #include "prototypes.h"
 #include "defines.h"
 #include <stdio.h>
 #include <stdlib.h>
 #include <ctype.h>
+#include <errno.h>
 #include "getdef.h"
 /*
  * A configuration item definition.
  */
 struct itemdef {
-	const char *name;	/* name of the item                     */
-	char *value;		/* value given, or NULL if no value     */
+	/*@null@*/const char *name;	/* name of the item                     */
+	/*@null@*/char *value;		/* value given, or NULL if no value     */
 };
 
 #define NUMDEFS	(sizeof(def_table)/sizeof(def_table[0]))
@@ -73,8 +77,10 @@ static struct itemdef def_table[] = {
 	{"PASS_MAX_DAYS", NULL},
 	{"PASS_MIN_DAYS", NULL},
 	{"PASS_WARN_AGE", NULL},
+#ifdef USE_SHA_CRYPT
 	{"SHA_CRYPT_MAX_ROUNDS", NULL},
 	{"SHA_CRYPT_MIN_ROUNDS", NULL},
+#endif
 	{"SULOG_FILE", NULL},
 	{"SU_NAME", NULL},
 	{"SYS_GID_MAX", NULL},
@@ -117,6 +123,11 @@ static struct itemdef def_table[] = {
 #ifdef USE_SYSLOG
 	{"SYSLOG_SG_ENAB", NULL},
 	{"SYSLOG_SU_ENAB", NULL},
+#endif
+#ifdef WITH_TCB
+	{"TCB_AUTH_GROUP", NULL},
+	{"TCB_SYMLINKS", NULL},
+	{"USE_TCB", NULL},
 #endif
 	{NULL, NULL}
 };
@@ -126,10 +137,10 @@ static struct itemdef def_table[] = {
 #endif
 
 static char def_fname[] = LOGINDEFS;	/* login config defs file       */
-static int def_loaded = 0;	/* are defs already loaded?     */
+static bool def_loaded = false;		/* are defs already loaded?     */
 
 /* local function prototypes */
-static struct itemdef *def_find (const char *);
+static /*@observer@*/ /*@null@*/struct itemdef *def_find (const char *);
 static void def_load (void);
 
 
@@ -140,14 +151,16 @@ static void def_load (void);
  * defined.  First time invoked, will load definitions from the file.
  */
 
-char *getdef_str (const char *item)
+/*@observer@*/ /*@null@*/const char *getdef_str (const char *item)
 {
 	struct itemdef *d;
 
-	if (!def_loaded)
+	if (!def_loaded) {
 		def_load ();
+	}
 
-	return ((d = def_find (item)) == NULL ? (char *) NULL : d->value);
+	d = def_find (item);
+	return ((NULL == d)? (const char *) NULL : d->value);
 }
 
 
@@ -157,15 +170,18 @@ char *getdef_str (const char *item)
  * Return TRUE if specified item is defined as "yes", else FALSE.
  */
 
-int getdef_bool (const char *item)
+bool getdef_bool (const char *item)
 {
 	struct itemdef *d;
 
-	if (!def_loaded)
+	if (!def_loaded) {
 		def_load ();
+	}
 
-	if ((d = def_find (item)) == NULL || d->value == NULL)
-		return 0;
+	d = def_find (item);
+	if ((NULL == d) || (NULL == d->value)) {
+		return false;
+	}
 
 	return (strcasecmp (d->value, "yes") == 0);
 }
@@ -182,14 +198,27 @@ int getdef_bool (const char *item)
 int getdef_num (const char *item, int dflt)
 {
 	struct itemdef *d;
+	long val;
 
-	if (!def_loaded)
+	if (!def_loaded) {
 		def_load ();
+	}
 
-	if ((d = def_find (item)) == NULL || d->value == NULL)
+	d = def_find (item);
+	if ((NULL == d) || (NULL == d->value)) {
 		return dflt;
+	}
 
-	return (int) strtol (d->value, (char **) NULL, 0);
+	if (   (getlong (d->value, &val) == 0)
+	    || (val > INT_MAX)
+	    || (val < INT_MIN)) {
+		fprintf (stderr,
+		         _("configuration error - cannot parse %s value: '%s'"),
+		         item, d->value);
+		return dflt;
+	}
+
+	return (int) val;
 }
 
 
@@ -204,14 +233,27 @@ int getdef_num (const char *item, int dflt)
 unsigned int getdef_unum (const char *item, unsigned int dflt)
 {
 	struct itemdef *d;
+	long val;
 
-	if (!def_loaded)
+	if (!def_loaded) {
 		def_load ();
+	}
 
-	if ((d = def_find (item)) == NULL || d->value == NULL)
+	d = def_find (item);
+	if ((NULL == d) || (NULL == d->value)) {
 		return dflt;
+	}
 
-	return (unsigned int) strtoul (d->value, (char **) NULL, 0);
+	if (   (getlong (d->value, &val) == 0)
+	    || (val < 0)
+	    || (val > INT_MAX)) {
+		fprintf (stderr,
+		         _("configuration error - cannot parse %s value: '%s'"),
+		         item, d->value);
+		return dflt;
+	}
+
+	return (unsigned int) val;
 }
 
 
@@ -226,16 +268,59 @@ unsigned int getdef_unum (const char *item, unsigned int dflt)
 long getdef_long (const char *item, long dflt)
 {
 	struct itemdef *d;
+	long val;
 
-	if (!def_loaded)
+	if (!def_loaded) {
 		def_load ();
+	}
 
-	if ((d = def_find (item)) == NULL || d->value == NULL)
+	d = def_find (item);
+	if ((NULL == d) || (NULL == d->value)) {
 		return dflt;
+	}
 
-	return strtol (d->value, (char **) NULL, 0);
+	if (getlong (d->value, &val) == 0) {
+		fprintf (stderr,
+		         _("configuration error - cannot parse %s value: '%s'"),
+		         item, d->value);
+		return dflt;
+	}
+
+	return val;
 }
 
+/*
+ * getdef_ulong - get unsigned long numerical value from table of definitions
+ *
+ * Returns numeric value of specified item, else the "dflt" value if
+ * the item is not defined.  Octal (leading "0") and hex (leading "0x")
+ * values are handled.
+ */
+
+unsigned long getdef_ulong (const char *item, unsigned long dflt)
+{
+	struct itemdef *d;
+	unsigned long val;
+
+	if (!def_loaded) {
+		def_load ();
+	}
+
+	d = def_find (item);
+	if ((NULL == d) || (NULL == d->value)) {
+		return dflt;
+	}
+
+	if (getulong (d->value, &val) == 0) {
+		/* FIXME: we should have a getulong */
+		fprintf (stderr,
+		         _("configuration error - cannot parse %s value: '%s'"),
+		         item, d->value);
+		return dflt;
+	}
+
+	return val;
+}
 
 /*
  * putdef_str - override the value read from /etc/login.defs
@@ -247,28 +332,33 @@ int putdef_str (const char *name, const char *value)
 	struct itemdef *d;
 	char *cp;
 
-	if (!def_loaded)
+	if (!def_loaded) {
 		def_load ();
+	}
 
 	/*
 	 * Locate the slot to save the value.  If this parameter
 	 * is unknown then "def_find" will print an err message.
 	 */
-	if ((d = def_find (name)) == NULL)
+	d = def_find (name);
+	if (NULL == d) {
 		return -1;
+	}
 
 	/*
 	 * Save off the value.
 	 */
-	if ((cp = strdup (value)) == NULL) {
-		fputs (_("Could not allocate space for config info.\n"),
-		       stderr);
+	cp = strdup (value);
+	if (NULL == cp) {
+		(void) fputs (_("Could not allocate space for config info.\n"),
+		              stderr);
 		SYSLOG ((LOG_ERR, "could not allocate space for config info"));
 		return -1;
 	}
 
-	if (d->value)
+	if (NULL != d->value) {
 		free (d->value);
+	}
 
 	d->value = cp;
 	return 0;
@@ -282,9 +372,8 @@ int putdef_str (const char *name, const char *value)
  * specified configuration option.
  */
 
-static struct itemdef *def_find (const char *name)
+static /*@observer@*/ /*@null@*/struct itemdef *def_find (const char *name)
 {
-	int n;
 	struct itemdef *ptr;
 
 
@@ -292,9 +381,10 @@ static struct itemdef *def_find (const char *name)
 	 * Search into the table.
 	 */
 
-	for (ptr = def_table; ptr->name; ptr++) {
-		if (!(n = strcmp (ptr->name, name)))
+	for (ptr = def_table; NULL != ptr->name; ptr++) {
+		if (strcmp (ptr->name, name) == 0) {
 			return ptr;
+		}
 	}
 
 	/*
@@ -302,9 +392,8 @@ static struct itemdef *def_find (const char *name)
 	 */
 
 	fprintf (stderr,
-		 _
-		 ("configuration error - unknown item '%s' (notify administrator)\n"),
-		 name);
+	         _("configuration error - unknown item '%s' (notify administrator)\n"),
+	         name);
 	SYSLOG ((LOG_CRIT, "unknown configuration item `%s'", name));
 	return (struct itemdef *) NULL;
 }
@@ -324,31 +413,35 @@ static void def_load (void)
 	/*
 	 * Open the configuration definitions file.
 	 */
-	if ((fp = fopen (def_fname, "r")) == NULL) {
-		SYSLOG ((LOG_CRIT, "cannot open login definitions %s [%m]",
-			 def_fname));
-		exit (1);
+	fp = fopen (def_fname, "r");
+	if (NULL == fp) {
+		int err = errno;
+		SYSLOG ((LOG_CRIT, "cannot open login definitions %s [%s]",
+		         def_fname, strerror (err)));
+		exit (EXIT_FAILURE);
 	}
 
 	/*
 	 * Set the initialized flag.
 	 * (do it early to prevent recursion in putdef_str())
 	 */
-	++def_loaded;
+	def_loaded = true;
 
 	/*
 	 * Go through all of the lines in the file.
 	 */
-	while (fgets (buf, sizeof (buf), fp) != NULL) {
+	while (fgets (buf, (int) sizeof (buf), fp) != NULL) {
 
 		/*
 		 * Trim trailing whitespace.
 		 */
-		for (i = strlen (buf) - 1; i >= 0; --i) {
-			if (!isspace (buf[i]))
+		for (i = (int) strlen (buf) - 1; i >= 0; --i) {
+			if (!isspace (buf[i])) {
 				break;
+			}
 		}
-		buf[++i] = '\0';
+		i++;
+		buf[i] = '\0';
 
 		/*
 		 * Break the line into two fields.
@@ -367,14 +460,19 @@ static void def_load (void)
 
 		/*
 		 * Store the value in def_table.
+		 *
+		 * Ignore failures to load the login.defs file.
+		 * The error was already reported to the user and to
+		 * syslog. The tools will just use their default values.
 		 */
-		putdef_str (name, value);
+		(void)putdef_str (name, value);
 	}
 
-	if (ferror (fp)) {
-		SYSLOG ((LOG_CRIT, "cannot read login definitions %s [%m]",
-			 def_fname));
-		exit (1);
+	if (ferror (fp) != 0) {
+		int err = errno;
+		SYSLOG ((LOG_CRIT, "cannot read login definitions %s [%s]",
+		         def_fname, strerror (err)));
+		exit (EXIT_FAILURE);
 	}
 
 	(void) fclose (fp);
@@ -391,18 +489,22 @@ int main (int argc, char **argv)
 	def_load ();
 
 	for (i = 0; i < NUMDEFS; ++i) {
-		if ((d = def_find (def_table[i].name)) == NULL)
+		d = def_find (def_table[i].name);
+		if (NULL == d) {
 			printf ("error - lookup '%s' failed\n",
-				def_table[i].name);
-		else
+			        def_table[i].name);
+		} else {
 			printf ("%4d %-24s %s\n", i + 1, d->name, d->value);
+		}
 	}
 	for (i = 1; i < argc; i++) {
-		if ((cp = getdef_str (argv[1])) != NULL)
+		cp = getdef_str (argv[1]);
+		if (NULL != cp) {
 			printf ("%s `%s'\n", argv[1], cp);
-		else
+		} else {
 			printf ("%s not found\n", argv[1]);
+		}
 	}
-	exit (0);
+	exit (EXIT_SUCCESS);
 }
 #endif
