@@ -1,8 +1,41 @@
+/*
+ * Copyright (c) 1990 - 1994, Julianne Frances Haugh
+ * Copyright (c) 1996 - 2000, Marek Michałkiewicz
+ * Copyright (c) 2001       , Michał Moskal
+ * Copyright (c) 2005       , Tomasz Kłoczko
+ * Copyright (c) 2007 - 2010, Nicolas François
+ * All rights reserved.
+ *
+ * Redistribution and use in source and binary forms, with or without
+ * modification, are permitted provided that the following conditions
+ * are met:
+ * 1. Redistributions of source code must retain the above copyright
+ *    notice, this list of conditions and the following disclaimer.
+ * 2. Redistributions in binary form must reproduce the above copyright
+ *    notice, this list of conditions and the following disclaimer in the
+ *    documentation and/or other materials provided with the distribution.
+ * 3. The name of the copyright holders or contributors may not be used to
+ *    endorse or promote products derived from this software without
+ *    specific prior written permission.
+ *
+ * THIS SOFTWARE IS PROVIDED BY THE COPYRIGHT HOLDERS AND CONTRIBUTORS
+ * ``AS IS'' AND ANY EXPRESS OR IMPLIED WARRANTIES, INCLUDING, BUT NOT
+ * LIMITED TO, THE IMPLIED WARRANTIES OF MERCHANTABILITY AND FITNESS FOR A
+ * PARTICULAR PURPOSE ARE DISCLAIMED.  IN NO EVENT SHALL THE COPYRIGHT
+ * HOLDERS OR CONTRIBUTORS BE LIABLE FOR ANY DIRECT, INDIRECT, INCIDENTAL,
+ * SPECIAL, EXEMPLARY, OR CONSEQUENTIAL DAMAGES (INCLUDING, BUT NOT
+ * LIMITED TO, PROCUREMENT OF SUBSTITUTE GOODS OR SERVICES; LOSS OF USE,
+ * DATA, OR PROFITS; OR BUSINESS INTERRUPTION) HOWEVER CAUSED AND ON ANY
+ * THEORY OF LIABILITY, WHETHER IN CONTRACT, STRICT LIABILITY, OR TORT
+ * (INCLUDING NEGLIGENCE OR OTHERWISE) ARISING IN ANY WAY OUT OF THE USE
+ * OF THIS SOFTWARE, EVEN IF ADVISED OF THE POSSIBILITY OF SUCH DAMAGE.
+ */
 
 #include <config.h>
 
-#ident "$Id: groupio.c 1753 2008-02-03 16:51:08Z nekral-guest $"
+#ident "$Id: groupio.c 3296 2011-02-16 20:32:16Z nekral-guest $"
 
+#include <assert.h>
 #include <stdio.h>
 
 #include "prototypes.h"
@@ -11,29 +44,24 @@
 #include "getdef.h"
 #include "groupio.h"
 
-static struct commonio_entry *merge_group_entries (struct commonio_entry *gr1,
-                                                   struct commonio_entry *gr2);
+static /*@null@*/struct commonio_entry *merge_group_entries (
+	/*@null@*/ /*@returned@*/struct commonio_entry *gr1,
+	/*@null@*/struct commonio_entry *gr2);
 static int split_groups (unsigned int max_members);
 static int group_open_hook (void);
 
-static void *group_dup (const void *ent)
+static /*@null@*/ /*@only@*/void *group_dup (const void *ent)
 {
 	const struct group *gr = ent;
 
 	return __gr_dup (gr);
 }
 
-static void group_free (void *ent)
+static void group_free (/*@out@*/ /*@only@*/void *ent)
 {
 	struct group *gr = ent;
 
-	free (gr->gr_name);
-	free (gr->gr_passwd);
-	while (*(gr->gr_mem)) {
-		free (*(gr->gr_mem));
-		gr->gr_mem++;
-	}
-	free (gr);
+	gr_free (gr);
 }
 
 static const char *group_getname (const void *ent)
@@ -52,6 +80,23 @@ static int group_put (const void *ent, FILE * file)
 {
 	const struct group *gr = ent;
 
+	if (   (NULL == gr)
+	    || (valid_field (gr->gr_name, ":\n") == -1)
+	    || (valid_field (gr->gr_passwd, ":\n") == -1)
+	    || (gr->gr_gid == (gid_t)-1)) {
+		return -1;
+	}
+
+	/* FIXME: fail also if gr->gr_mem == NULL ?*/
+	if (NULL != gr->gr_mem) {
+		size_t i;
+		for (i = 0; NULL != gr->gr_mem[i]; i++) {
+			if (valid_field (gr->gr_mem[i], ",:\n") == -1) {
+				return -1;
+			}
+		}
+	}
+
 	return (putgrent (gr, file) == -1) ? -1 : 0;
 }
 
@@ -59,8 +104,9 @@ static int group_close_hook (void)
 {
 	unsigned int max_members = getdef_unum("MAX_MEMBERS_PER_GROUP", 0);
 
-	if (0 == max_members)
+	if (0 == max_members) {
 		return 1;
+	}
 
 	return split_groups (max_members);
 }
@@ -77,7 +123,7 @@ static struct commonio_ops group_ops = {
 	group_close_hook
 };
 
-static struct commonio_db group_db = {
+static /*@owned@*/struct commonio_db group_db = {
 	GROUP_FILE,		/* filename */
 	&group_ops,		/* ops */
 	NULL,			/* fp */
@@ -87,17 +133,22 @@ static struct commonio_db group_db = {
 	NULL,			/* head */
 	NULL,			/* tail */
 	NULL,			/* cursor */
-	0,			/* changed */
-	0,			/* isopen */
-	0,			/* locked */
-	0			/* readonly */
+	false,			/* changed */
+	false,			/* isopen */
+	false,			/* locked */
+	false			/* readonly */
 };
 
-int gr_name (const char *filename)
+int gr_setdbname (const char *filename)
 {
 	return commonio_setname (&group_db, filename);
 }
 
+/*@observer@*/const char *gr_dbname (void)
+{
+	return group_db.filename;
+}
+
 int gr_lock (void)
 {
 	return commonio_lock (&group_db);
@@ -108,12 +159,12 @@ int gr_open (int mode)
 	return commonio_open (&group_db, mode);
 }
 
-const struct group *gr_locate (const char *name)
+/*@observer@*/ /*@null@*/const struct group *gr_locate (const char *name)
 {
 	return commonio_locate (&group_db, name);
 }
 
-const struct group *gr_locate_gid (gid_t gid)
+/*@observer@*/ /*@null@*/const struct group *gr_locate_gid (gid_t gid)
 {
 	const struct group *grp;
 
@@ -140,7 +191,7 @@ int gr_rewind (void)
 	return commonio_rewind (&group_db);
 }
 
-const struct group *gr_next (void)
+/*@observer@*/ /*@null@*/const struct group *gr_next (void)
 {
 	return commonio_next (&group_db);
 }
@@ -157,15 +208,15 @@ int gr_unlock (void)
 
 void __gr_set_changed (void)
 {
-	group_db.changed = 1;
+	group_db.changed = true;
 }
 
-struct commonio_entry *__gr_get_head (void)
+/*@dependent@*/ /*@null@*/struct commonio_entry *__gr_get_head (void)
 {
 	return group_db.head;
 }
 
-struct commonio_db *__gr_get_db (void)
+/*@observer@*/const struct commonio_db *__gr_get_db (void)
 {
 	return &group_db;
 }
@@ -179,20 +230,23 @@ static int gr_cmp (const void *p1, const void *p2)
 {
 	gid_t u1, u2;
 
-	if ((*(struct commonio_entry **) p1)->eptr == NULL)
+	if ((*(struct commonio_entry **) p1)->eptr == NULL) {
 		return 1;
-	if ((*(struct commonio_entry **) p2)->eptr == NULL)
+	}
+	if ((*(struct commonio_entry **) p2)->eptr == NULL) {
 		return -1;
+	}
 
 	u1 = ((struct group *) (*(struct commonio_entry **) p1)->eptr)->gr_gid;
 	u2 = ((struct group *) (*(struct commonio_entry **) p2)->eptr)->gr_gid;
 
-	if (u1 < u2)
+	if (u1 < u2) {
 		return -1;
-	else if (u1 > u2)
+	} else if (u1 > u2) {
 		return 1;
-	else
+	} else {
 		return 0;
+	}
 }
 
 /* Sort entries by GID */
@@ -206,11 +260,12 @@ static int group_open_hook (void)
 	unsigned int max_members = getdef_unum("MAX_MEMBERS_PER_GROUP", 0);
 	struct commonio_entry *gr1, *gr2;
 
-	if (0 == max_members)
+	if (0 == max_members) {
 		return 1;
+	}
 
-	for (gr1 = group_db.head; gr1; gr1 = gr1->next) {
-		for (gr2 = gr1->next; gr2; gr2 = gr2->next) {
+	for (gr1 = group_db.head; NULL != gr1; gr1 = gr1->next) {
+		for (gr2 = gr1->next; NULL != gr2; gr2 = gr2->next) {
 			struct group *g1 = (struct group *)gr1->eptr;
 			struct group *g2 = (struct group *)gr2->eptr;
 			if (NULL != g1 &&
@@ -225,11 +280,15 @@ static int group_open_hook (void)
 				if (NULL == gr1)
 					return 0;
 				/* Unlink gr2 */
-				if (NULL != gr2->next)
+				if (NULL != gr2->next) {
 					gr2->next->prev = gr2->prev;
+				}
+				/* gr2 does not start with head */
+				assert (NULL != gr2->prev);
 				gr2->prev->next = gr2->next;
 			}
 		}
+		assert (NULL != gr1);
 	}
 
 	return 1;
@@ -247,15 +306,16 @@ static int group_open_hook (void)
  * the modified first entry on success, or NULL on failure (with errno
  * set).
  */
-static struct commonio_entry *merge_group_entries (struct commonio_entry *gr1,
-                                                   struct commonio_entry *gr2)
+static /*@null@*/struct commonio_entry *merge_group_entries (
+	/*@null@*/ /*@returned@*/struct commonio_entry *gr1,
+	/*@null@*/struct commonio_entry *gr2)
 {
 	struct group *gptr1;
 	struct group *gptr2;
 	char **new_members;
-	int members = 0;
+	size_t members = 0;
 	char *new_line;
-	int new_line_len, i;
+	size_t new_line_len, i;
 	if (NULL == gr2 || NULL == gr1) {
 		errno = EINVAL;
 		return NULL;
@@ -284,30 +344,36 @@ static struct commonio_entry *merge_group_entries (struct commonio_entry *gr1,
 	for (i=0; NULL != gptr2->gr_mem[i]; i++) {
 		char **pmember = gptr1->gr_mem;
 		while (NULL != *pmember) {
-			if (0 == strcmp(*pmember, gptr2->gr_mem[i]))
+			if (0 == strcmp(*pmember, gptr2->gr_mem[i])) {
 				break;
+			}
 			pmember++;
 		}
-		if (NULL == *pmember)
+		if (NULL == *pmember) {
 			members++;
+		}
 	}
 	new_members = (char **)malloc ( (members+1) * sizeof(char*) );
 	if (NULL == new_members) {
+		free (new_line);
 		errno = ENOMEM;
 		return NULL;
 	}
-	for (i=0; NULL != gptr1->gr_mem[i]; i++)
+	for (i=0; NULL != gptr1->gr_mem[i]; i++) {
 		new_members[i] = gptr1->gr_mem[i];
+	}
 	members = i;
 	for (i=0; NULL != gptr2->gr_mem[i]; i++) {
 		char **pmember = new_members;
 		while (NULL != *pmember) {
-			if (0 == strcmp(*pmember, gptr2->gr_mem[i]))
+			if (0 == strcmp(*pmember, gptr2->gr_mem[i])) {
 				break;
+			}
 			pmember++;
 		}
 		if (NULL == *pmember) {
-			new_members[members++] = gptr2->gr_mem[i];
+			new_members[members] = gptr2->gr_mem[i];
+			members++;
 			new_members[members] = NULL;
 		}
 	}
@@ -328,20 +394,24 @@ static int split_groups (unsigned int max_members)
 {
 	struct commonio_entry *gr;
 
-	for (gr = group_db.head; gr; gr = gr->next) {
+	for (gr = group_db.head; NULL != gr; gr = gr->next) {
 		struct group *gptr = (struct group *)gr->eptr;
 		struct commonio_entry *new;
 		struct group *new_gptr;
 		unsigned int members = 0;
+		unsigned int i;
 
 		/* Check if this group must be split */
-		if (!gr->changed)
+		if (!gr->changed) {
 			continue;
-		if (NULL == gptr)
+		}
+		if (NULL == gptr) {
 			continue;
+		}
 		for (members = 0; NULL != gptr->gr_mem[members]; members++);
-		if (members <= max_members)
+		if (members <= max_members) {
 			continue;
+		}
 
 		new = (struct commonio_entry *) malloc (sizeof *new);
 		if (NULL == new) {
@@ -350,17 +420,32 @@ static int split_groups (unsigned int max_members)
 		}
 		new->eptr = group_dup(gr->eptr);
 		if (NULL == new->eptr) {
+			free (new);
 			errno = ENOMEM;
 			return 0;
 		}
 		new_gptr = (struct group *)new->eptr;
 		new->line = NULL;
-		new->changed = 1;
+		new->changed = true;
 
 		/* Enforce the maximum number of members on gptr */
-		gptr->gr_mem[max_members] = NULL;
+		for (i = max_members; NULL != gptr->gr_mem[i]; i++) {
+			free (gptr->gr_mem[i]);
+			gptr->gr_mem[i] = NULL;
+		}
+		/* Shift all the members */
 		/* The number of members in new_gptr will be check later */
-		new_gptr->gr_mem = &new_gptr->gr_mem[max_members];
+		for (i = 0; NULL != new_gptr->gr_mem[i + max_members]; i++) {
+			if (NULL != new_gptr->gr_mem[i]) {
+				free (new_gptr->gr_mem[i]);
+			}
+			new_gptr->gr_mem[i] = new_gptr->gr_mem[i + max_members];
+			new_gptr->gr_mem[i + max_members] = NULL;
+		}
+		for (; NULL != new_gptr->gr_mem[i]; i++) {
+			free (new_gptr->gr_mem[i]);
+			new_gptr->gr_mem[i] = NULL;
+		}
 
 		/* insert the new entry in the list */
 		new->prev = gr;
