@@ -1,5 +1,8 @@
 /*
- * Copyright 1989 - 1994, Julianne Frances Haugh
+ * Copyright (c) 1989 - 1994, Julianne Frances Haugh
+ * Copyright (c) 1996 - 1999, Marek Michałkiewicz
+ * Copyright (c) 2003 - 2005, Tomasz Kłoczko
+ * Copyright (c) 2007 - 2010, Nicolas François
  * All rights reserved.
  *
  * Redistribution and use in source and binary forms, with or without
@@ -10,28 +13,28 @@
  * 2. Redistributions in binary form must reproduce the above copyright
  *    notice, this list of conditions and the following disclaimer in the
  *    documentation and/or other materials provided with the distribution.
- * 3. Neither the name of Julianne F. Haugh nor the names of its contributors
- *    may be used to endorse or promote products derived from this software
- *    without specific prior written permission.
+ * 3. The name of the copyright holders or contributors may not be used to
+ *    endorse or promote products derived from this software without
+ *    specific prior written permission.
  *
- * THIS SOFTWARE IS PROVIDED BY JULIE HAUGH AND CONTRIBUTORS ``AS IS'' AND
- * ANY EXPRESS OR IMPLIED WARRANTIES, INCLUDING, BUT NOT LIMITED TO, THE
- * IMPLIED WARRANTIES OF MERCHANTABILITY AND FITNESS FOR A PARTICULAR PURPOSE
- * ARE DISCLAIMED.  IN NO EVENT SHALL JULIE HAUGH OR CONTRIBUTORS BE LIABLE
- * FOR ANY DIRECT, INDIRECT, INCIDENTAL, SPECIAL, EXEMPLARY, OR CONSEQUENTIAL
- * DAMAGES (INCLUDING, BUT NOT LIMITED TO, PROCUREMENT OF SUBSTITUTE GOODS
- * OR SERVICES; LOSS OF USE, DATA, OR PROFITS; OR BUSINESS INTERRUPTION)
- * HOWEVER CAUSED AND ON ANY THEORY OF LIABILITY, WHETHER IN CONTRACT, STRICT
- * LIABILITY, OR TORT (INCLUDING NEGLIGENCE OR OTHERWISE) ARISING IN ANY WAY
- * OUT OF THE USE OF THIS SOFTWARE, EVEN IF ADVISED OF THE POSSIBILITY OF
- * SUCH DAMAGE.
+ * THIS SOFTWARE IS PROVIDED BY THE COPYRIGHT HOLDERS AND CONTRIBUTORS
+ * ``AS IS'' AND ANY EXPRESS OR IMPLIED WARRANTIES, INCLUDING, BUT NOT
+ * LIMITED TO, THE IMPLIED WARRANTIES OF MERCHANTABILITY AND FITNESS FOR A
+ * PARTICULAR PURPOSE ARE DISCLAIMED.  IN NO EVENT SHALL THE COPYRIGHT
+ * HOLDERS OR CONTRIBUTORS BE LIABLE FOR ANY DIRECT, INDIRECT, INCIDENTAL,
+ * SPECIAL, EXEMPLARY, OR CONSEQUENTIAL DAMAGES (INCLUDING, BUT NOT
+ * LIMITED TO, PROCUREMENT OF SUBSTITUTE GOODS OR SERVICES; LOSS OF USE,
+ * DATA, OR PROFITS; OR BUSINESS INTERRUPTION) HOWEVER CAUSED AND ON ANY
+ * THEORY OF LIABILITY, WHETHER IN CONTRACT, STRICT LIABILITY, OR TORT
+ * (INCLUDING NEGLIGENCE OR OTHERWISE) ARISING IN ANY WAY OUT OF THE USE
+ * OF THIS SOFTWARE, EVEN IF ADVISED OF THE POSSIBILITY OF SUCH DAMAGE.
  */
 
 #include <config.h>
 
 #ifndef USE_PAM
 
-#ident "$Id: obscure.c 1668 2008-01-06 13:20:25Z nekral-guest $"
+#ident "$Id: obscure.c 3232 2010-08-22 19:13:53Z nekral-guest $"
 
 
 /*
@@ -47,24 +50,26 @@
 /*
  * can't be a palindrome - like `R A D A R' or `M A D A M'
  */
-static int palindrome (unused const char *old, const char *new)
+static bool palindrome (unused const char *old, const char *new)
 {
-	int i, j;
+	size_t i, j;
 
 	i = strlen (new);
 
-	for (j = 0; j < i; j++)
-		if (new[i - j - 1] != new[j])
-			return 0;
+	for (j = 0; j < i; j++) {
+		if (new[i - j - 1] != new[j]) {
+			return false;
+		}
+	}
 
-	return 1;
+	return true;
 }
 
 /*
  * more than half of the characters are different ones.
  */
 
-static int similar (const char *old, const char *new)
+static bool similar (/*@notnull@*/const char *old, /*@notnull@*/const char *new)
 {
 	int i, j;
 
@@ -74,41 +79,46 @@ static int similar (const char *old, const char *new)
 	 * the new password is long enough.  Please feel free to suggest
 	 * something better...  --marekm
 	 */
-	if (strlen (new) >= 8)
-		return 0;
+	if (strlen (new) >= 8) {
+		return false;
+	}
 
-	for (i = j = 0; new[i] && old[i]; i++)
-		if (strchr (new, old[i]))
+	for (i = j = 0; ('\0' != new[i]) && ('\0' != old[i]); i++) {
+		if (strchr (new, old[i]) != NULL) {
 			j++;
+		}
+	}
 
-	if (i >= j * 2)
-		return 0;
+	if (i >= j * 2) {
+		return false;
+	}
 
-	return 1;
+	return true;
 }
 
 /*
  * a nice mix of characters.
  */
 
-static int simple (unused const char *old, const char *new)
+static bool simple (unused const char *old, const char *new)
 {
-	int digits = 0;
-	int uppers = 0;
-	int lowers = 0;
-	int others = 0;
+	bool digits = false;
+	bool uppers = false;
+	bool lowers = false;
+	bool others = false;
 	int size;
 	int i;
 
-	for (i = 0; new[i]; i++) {
-		if (isdigit (new[i]))
-			digits++;
-		else if (isupper (new[i]))
-			uppers++;
-		else if (islower (new[i]))
-			lowers++;
-		else
-			others++;
+	for (i = 0; '\0' != new[i]; i++) {
+		if (isdigit (new[i])) {
+			digits = true;
+		} else if (isupper (new[i])) {
+			uppers = true;
+		} else if (islower (new[i])) {
+			lowers = true;
+		} else {
+			others = true;
+		}
 	}
 
 	/*
@@ -117,32 +127,40 @@ static int simple (unused const char *old, const char *new)
 	 */
 
 	size = 9;
-	if (digits)
+	if (digits) {
 		size--;
-	if (uppers)
+	}
+	if (uppers) {
 		size--;
-	if (lowers)
+	}
+	if (lowers) {
 		size--;
-	if (others)
+	}
+	if (others) {
 		size--;
+	}
 
-	if (size <= i)
-		return 0;
+	if (size <= i) {
+		return false;
+	}
 
-	return 1;
+	return true;
 }
 
-static char *str_lower (char *string)
+static char *str_lower (/*@returned@*/char *string)
 {
 	char *cp;
 
-	for (cp = string; *cp; cp++)
+	for (cp = string; '\0' != *cp; cp++) {
 		*cp = tolower (*cp);
+	}
 	return string;
 }
 
-static const char *password_check (const char *old, const char *new,
-				   const struct passwd *pwdp)
+static /*@observer@*//*@null@*/const char *password_check (
+	/*@notnull@*/const char *old,
+	/*@notnull@*/const char *new,
+	/*@notnull@*/const struct passwd *pwdp)
 {
 	const char *msg = NULL;
 	char *oldmono, *newmono, *wrapped;
@@ -157,8 +175,9 @@ static const char *password_check (const char *old, const char *new,
 #endif
 #endif
 
-	if (strcmp (new, old) == 0)
+	if (strcmp (new, old) == 0) {
 		return _("no change");
+	}
 
 	newmono = str_lower (xstrdup (new));
 	oldmono = str_lower (xstrdup (old));
@@ -166,33 +185,32 @@ static const char *password_check (const char *old, const char *new,
 	strcpy (wrapped, oldmono);
 	strcat (wrapped, oldmono);
 
-	if (palindrome (oldmono, newmono))
+	if (palindrome (oldmono, newmono)) {
 		msg = _("a palindrome");
-
-	if (!msg && strcmp (oldmono, newmono) == 0)
+	} else if (strcmp (oldmono, newmono) == 0) {
 		msg = _("case changes only");
-
-	if (!msg && similar (oldmono, newmono))
+	} else if (similar (oldmono, newmono)) {
 		msg = _("too similar");
-
-	if (!msg && simple (old, new))
+	} else if (simple (old, new)) {
 		msg = _("too simple");
-
-	if (!msg && strstr (wrapped, newmono))
+	} else if (strstr (wrapped, newmono) != NULL) {
 		msg = _("rotated");
-
+	} else {
 #ifdef HAVE_LIBCRACK
-	/*
-	 * Invoke Alec Muffett's cracklib routines.
-	 */
+		/*
+		 * Invoke Alec Muffett's cracklib routines.
+		 */
 
-	if (!msg && (dictpath = getdef_str ("CRACKLIB_DICTPATH")))
+		dictpath = getdef_str ("CRACKLIB_DICTPATH");
+		if (NULL != dictpath) {
 #ifdef HAVE_LIBCRACK_PW
-		msg = FascistCheckPw (new, dictpath, pwdp);
+			msg = FascistCheckPw (new, dictpath, pwdp);
 #else
-		msg = FascistCheck (new, dictpath);
+			msg = FascistCheck (new, dictpath);
 #endif
+		}
 #endif
+	}
 	strzero (newmono);
 	strzero (oldmono);
 	strzero (wrapped);
@@ -203,62 +221,73 @@ static const char *password_check (const char *old, const char *new,
 	return msg;
 }
 
- /*ARGSUSED*/
-    static const char *obscure_msg (const char *old, const char *new,
-				    const struct passwd *pwdp)
+static /*@observer@*//*@null@*/const char *obscure_msg (
+	/*@notnull@*/const char *old,
+	/*@notnull@*/const char *new,
+	/*@notnull@*/const struct passwd *pwdp)
 {
-	int maxlen, oldlen, newlen;
+	size_t maxlen, oldlen, newlen;
 	char *new1, *old1;
 	const char *msg;
-	char *result;
+	const char *result;
 
 	oldlen = strlen (old);
 	newlen = strlen (new);
 
-	if (newlen < getdef_num ("PASS_MIN_LEN", 0))
+	if (newlen < (size_t) getdef_num ("PASS_MIN_LEN", 0)) {
 		return _("too short");
+	}
 
 	/*
 	 * Remaining checks are optional.
 	 */
-	if (!getdef_bool ("OBSCURE_CHECKS_ENAB"))
+	if (!getdef_bool ("OBSCURE_CHECKS_ENAB")) {
 		return NULL;
+	}
 
 	msg = password_check (old, new, pwdp);
-	if (msg)
+	if (NULL != msg) {
 		return msg;
+	}
 
-	if ((result = getdef_str ("ENCRYPT_METHOD")) == NULL) {
+	result = getdef_str ("ENCRYPT_METHOD");
+	if (NULL == result) {
 	/* The traditional crypt() truncates passwords to 8 chars.  It is
 	   possible to circumvent the above checks by choosing an easy
 	   8-char password and adding some random characters to it...
 	   Example: "password$%^&*123".  So check it again, this time
 	   truncated to the maximum length.  Idea from npasswd.  --marekm */
 
-		if (getdef_bool ("MD5_CRYPT_ENAB"))
+		if (getdef_bool ("MD5_CRYPT_ENAB")) {
 			return NULL;
+		}
 
 	} else {
 
-		if (   !strcmp (result, "MD5")
+		if (   (strcmp (result, "MD5")    == 0)
 #ifdef USE_SHA_CRYPT
-		    || !strcmp (result, "SHA256")
-		    || !strcmp (result, "SHA512")
+		    || (strcmp (result, "SHA256") == 0)
+		    || (strcmp (result, "SHA512") == 0)
 #endif
-		    )
+		    ) {
 			return NULL;
+		}
 
 	}
-	maxlen = getdef_num ("PASS_MAX_LEN", 8);
-	if (oldlen <= maxlen && newlen <= maxlen)
+	maxlen = (size_t) getdef_num ("PASS_MAX_LEN", 8);
+	if (   (oldlen <= maxlen)
+	    && (newlen <= maxlen)) {
 		return NULL;
+	}
 
 	new1 = xstrdup (new);
 	old1 = xstrdup (old);
-	if (newlen > maxlen)
+	if (newlen > maxlen) {
 		new1[maxlen] = '\0';
-	if (oldlen > maxlen)
+	}
+	if (oldlen > maxlen) {
 		old1[maxlen] = '\0';
+	}
 
 	msg = password_check (old1, new1, pwdp);
 
@@ -278,15 +307,15 @@ static const char *password_check (const char *old, const char *new,
  *	check passwords.
  */
 
-int obscure (const char *old, const char *new, const struct passwd *pwdp)
+bool obscure (const char *old, const char *new, const struct passwd *pwdp)
 {
 	const char *msg = obscure_msg (old, new, pwdp);
 
-	if (msg) {
+	if (NULL != msg) {
 		printf (_("Bad password: %s.  "), msg);
-		return 0;
+		return false;
 	}
-	return 1;
+	return true;
 }
 
 #else				/* !USE_PAM */
