@@ -1,5 +1,8 @@
 /*
- * Copyright 1989 - 1994, Julianne Frances Haugh
+ * Copyright (c) 1989 - 1994, Julianne Frances Haugh
+ * Copyright (c) 1996 - 2000, Marek Michałkiewicz
+ * Copyright (c) 2001 - 2006, Tomasz Kłoczko
+ * Copyright (c) 2007 - 2011, Nicolas François
  * All rights reserved.
  *
  * Redistribution and use in source and binary forms, with or without
@@ -10,38 +13,38 @@
  * 2. Redistributions in binary form must reproduce the above copyright
  *    notice, this list of conditions and the following disclaimer in the
  *    documentation and/or other materials provided with the distribution.
- * 3. Neither the name of Julianne F. Haugh nor the names of its contributors
- *    may be used to endorse or promote products derived from this software
- *    without specific prior written permission.
+ * 3. The name of the copyright holders or contributors may not be used to
+ *    endorse or promote products derived from this software without
+ *    specific prior written permission.
  *
- * THIS SOFTWARE IS PROVIDED BY JULIE HAUGH AND CONTRIBUTORS ``AS IS'' AND
- * ANY EXPRESS OR IMPLIED WARRANTIES, INCLUDING, BUT NOT LIMITED TO, THE
- * IMPLIED WARRANTIES OF MERCHANTABILITY AND FITNESS FOR A PARTICULAR PURPOSE
- * ARE DISCLAIMED. IN NO EVENT SHALL JULIE HAUGH OR CONTRIBUTORS BE LIABLE
- * FOR ANY DIRECT, INDIRECT, INCIDENTAL, SPECIAL, EXEMPLARY, OR CONSEQUENTIAL
- * DAMAGES (INCLUDING, BUT NOT LIMITED TO, PROCUREMENT OF SUBSTITUTE GOODS
- * OR SERVICES; LOSS OF USE, DATA, OR PROFITS; OR BUSINESS INTERRUPTION)
- * HOWEVER CAUSED AND ON ANY THEORY OF LIABILITY, WHETHER IN CONTRACT, STRICT
- * LIABILITY, OR TORT (INCLUDING NEGLIGENCE OR OTHERWISE) ARISING IN ANY WAY
- * OUT OF THE USE OF THIS SOFTWARE, EVEN IF ADVISED OF THE POSSIBILITY OF
- * SUCH DAMAGE.
+ * THIS SOFTWARE IS PROVIDED BY THE COPYRIGHT HOLDERS AND CONTRIBUTORS
+ * ``AS IS'' AND ANY EXPRESS OR IMPLIED WARRANTIES, INCLUDING, BUT NOT
+ * LIMITED TO, THE IMPLIED WARRANTIES OF MERCHANTABILITY AND FITNESS FOR A
+ * PARTICULAR PURPOSE ARE DISCLAIMED.  IN NO EVENT SHALL THE COPYRIGHT
+ * HOLDERS OR CONTRIBUTORS BE LIABLE FOR ANY DIRECT, INDIRECT, INCIDENTAL,
+ * SPECIAL, EXEMPLARY, OR CONSEQUENTIAL DAMAGES (INCLUDING, BUT NOT
+ * LIMITED TO, PROCUREMENT OF SUBSTITUTE GOODS OR SERVICES; LOSS OF USE,
+ * DATA, OR PROFITS; OR BUSINESS INTERRUPTION) HOWEVER CAUSED AND ON ANY
+ * THEORY OF LIABILITY, WHETHER IN CONTRACT, STRICT LIABILITY, OR TORT
+ * (INCLUDING NEGLIGENCE OR OTHERWISE) ARISING IN ANY WAY OUT OF THE USE
+ * OF THIS SOFTWARE, EVEN IF ADVISED OF THE POSSIBILITY OF SUCH DAMAGE.
  */
 
 #include <config.h>
 
-#ident "$Id: chfn.c 1751 2008-02-03 16:28:03Z nekral-guest $"
+#ident "$Id: chfn.c 3576 2011-11-13 16:24:57Z nekral-guest $"
 
 #include <fcntl.h>
 #include <pwd.h>
 #include <signal.h>
 #include <stdio.h>
 #include <sys/types.h>
+#include <getopt.h>
 #ifdef WITH_SELINUX
 #include <selinux/selinux.h>
 #include <selinux/av_permissions.h>
 #endif
 #include "defines.h"
-#include "exitcodes.h"
 #include "getdef.h"
 #include "nscd.h"
 #ifdef USE_PAM
@@ -50,33 +53,35 @@
 #include "prototypes.h"
 #include "pwauth.h"
 #include "pwio.h"
+/*@-exitarg@*/
+#include "exitcodes.h"
+
 /*
  * Global variables.
  */
-static char *Prog;
+const char *Prog;
 static char fullnm[BUFSIZ];
 static char roomno[BUFSIZ];
 static char workph[BUFSIZ];
 static char homeph[BUFSIZ];
 static char slop[BUFSIZ];
-static int amroot;
+static bool amroot;
 /* Flags */
-static int fflg = 0;		/* -f - set full name                */
-static int rflg = 0;		/* -r - set room number              */
-static int wflg = 0;		/* -w - set work phone number        */
-static int hflg = 0;		/* -h - set home phone number        */
-static int oflg = 0;		/* -o - set other information        */
-#ifdef USE_PAM
-static pam_handle_t *pamh = NULL;
-#endif
+static bool fflg = false;		/* -f - set full name                */
+static bool rflg = false;		/* -r - set room number              */
+static bool wflg = false;		/* -w - set work phone number        */
+static bool hflg = false;		/* -h - set home phone number        */
+static bool oflg = false;		/* -o - set other information        */
+static bool pw_locked = false;
 
 /*
  * External identifiers
  */
 
 /* local function prototypes */
-static void usage (void);
-static int may_change_field (int);
+static void fail_exit (int code);
+static /*@noreturn@*/void usage (int status);
+static bool may_change_field (int);
 static void new_fields (void);
 static char *copy_field (char *, char *, char *);
 static void process_flags (int argc, char **argv);
@@ -84,22 +89,45 @@ static void check_perms (const struct passwd *pw);
 static void update_gecos (const char *user, char *gecos);
 static void get_old_fields (const char *gecos);
 
+/*
+ * fail_exit - exit with an error and do some cleanup
+ */
+static void fail_exit (int code)
+{
+	if (pw_locked) {
+		if (pw_unlock () == 0) {
+			fprintf (stderr, _("%s: failed to unlock %s\n"), Prog, pw_dbname ());
+			SYSLOG ((LOG_ERR, "failed to unlock %s", pw_dbname ()));
+			/* continue */
+		}
+	}
+	pw_locked = false;
+
+	closelog ();
+
+	exit (code);
+}
+
 /*
  * usage - print command line syntax and exit
  */
-static void usage (void)
+static /*@noreturn@*/void usage (int status)
 {
-	if (amroot) {
-		fprintf (stderr,
-		         _("Usage: %s [-f full_name] [-r room_no] "
-		           "[-w work_ph]\n"
-		           "\t[-h home_ph] [-o other] [user]\n"), Prog);
-	} else {
-		fprintf (stderr,
-		         _("Usage: %s [-f full_name] [-r room_no] "
-		           "[-w work_ph] [-h home_ph]\n"), Prog);
-	}
-	exit (E_USAGE);
+	FILE *usageout = (E_SUCCESS != status) ? stderr : stdout;
+	(void) fprintf (usageout,
+	                _("Usage: %s [options] [LOGIN]\n"
+	                  "\n"
+	                  "Options:\n"),
+	                Prog);
+	(void) fputs (_("  -f, --full-name FULL_NAME     change user's full name\n"), usageout);
+	(void) fputs (_("  -h, --home-phone HOME_PHONE   change user's home phone number\n"), usageout);
+	(void) fputs (_("  -o, --other OTHER_INFO        change user's other GECOS information\n"), usageout);
+	(void) fputs (_("  -r, --room ROOM_NUMBER        change user's room number\n"), usageout);
+	(void) fputs (_("  -R, --root CHROOT_DIR         directory to chroot into\n"), usageout);
+	(void) fputs (_("  -u, --help                    display this help message and exit\n"), usageout);
+	(void) fputs (_("  -w, --work-phone WORK_PHONE   change user's office phone number\n"), usageout);
+	(void) fputs ("\n", usageout);
+	exit (status);
 }
 
 /*
@@ -110,9 +138,9 @@ static void usage (void)
  *
  *	field should be one of 'f', 'r', 'w', 'h'
  *
- *	Return 1 if the user can change the field and 0 otherwise.
+ *	Return true if the user can change the field and false otherwise.
  */
-static int may_change_field (int field)
+static bool may_change_field (int field)
 {
 	const char *cp;
 
@@ -134,7 +162,7 @@ static int may_change_field (int field)
 	 * if the string is empty or not defined at all.
 	 */
 	if (amroot) {
-		return 1;
+		return true;
 	}
 
 	cp = getdef_str ("CHFN_RESTRICT");
@@ -147,10 +175,10 @@ static int may_change_field (int field)
 	}
 
 	if (strchr (cp, field) != NULL) {
-		return 1;
+		return true;
 	}
 
-	return 0;
+	return false;
 }
 
 /*
@@ -166,25 +194,25 @@ static void new_fields (void)
 	if (may_change_field ('f')) {
 		change_field (fullnm, sizeof fullnm, _("Full Name"));
 	} else {
-		printf (_("\tFull Name: %s\n"), fullnm);
+		printf (_("\t%s: %s\n"), _("Full Name"), fullnm);
 	}
 
 	if (may_change_field ('r')) {
 		change_field (roomno, sizeof roomno, _("Room Number"));
 	} else {
-		printf (_("\tRoom Number: %s\n"), roomno);
+		printf (_("\t%s: %s\n"), _("Room Number"), fullnm);
 	}
 
 	if (may_change_field ('w')) {
 		change_field (workph, sizeof workph, _("Work Phone"));
 	} else {
-		printf (_("\tWork Phone: %s\n"), workph);
+		printf (_("\t%s: %s\n"), _("Work Phone"), fullnm);
 	}
 
 	if (may_change_field ('h')) {
 		change_field (homeph, sizeof homeph, _("Home Phone"));
 	} else {
-		printf (_("\tHome Phone: %s\n"), homeph);
+		printf (_("\t%s: %s\n"), _("Home Phone"), fullnm);
 	}
 
 	if (amroot) {
@@ -239,7 +267,17 @@ static char *copy_field (char *in, char *out, char *extra)
  */
 static void process_flags (int argc, char **argv)
 {
-	int flag;		/* flag currently being processed    */
+	int c;		/* flag currently being processed    */
+	static struct option long_options[] = {
+		{"full-name",  required_argument, NULL, 'f'},
+		{"home-phone", required_argument, NULL, 'h'},
+		{"other",      required_argument, NULL, 'o'},
+		{"room",       required_argument, NULL, 'r'},
+		{"root",       required_argument, NULL, 'R'},
+		{"help",       no_argument,       NULL, 'u'},
+		{"work-phone", required_argument, NULL, 'w'},
+		{NULL, 0, NULL, '\0'}
+	};
 
 	/* 
 	 * The remaining arguments will be processed one by one and executed
@@ -248,15 +286,16 @@ static void process_flags (int argc, char **argv)
 	 * environment and must agree with the real UID. Also, the UID will
 	 * be checked for any commands which are restricted to root only.
 	 */
-	while ((flag = getopt (argc, argv, "f:r:w:h:o:")) != EOF) {
-		switch (flag) {
+	while ((c = getopt_long (argc, argv, "f:h:o:r:R:uw:",
+	                         long_options, NULL)) != -1) {
+		switch (c) {
 		case 'f':
 			if (!may_change_field ('f')) {
 				fprintf (stderr,
 				         _("%s: Permission denied.\n"), Prog);
 				exit (E_NOPERM);
 			}
-			fflg++;
+			fflg = true;
 			STRFCPY (fullnm, optarg);
 			break;
 		case 'h':
@@ -265,38 +304,43 @@ static void process_flags (int argc, char **argv)
 				         _("%s: Permission denied.\n"), Prog);
 				exit (E_NOPERM);
 			}
-			hflg++;
+			hflg = true;
 			STRFCPY (homeph, optarg);
 			break;
-		case 'r':
-			if (!may_change_field ('r')) {
-				fprintf (stderr,
-				         _("%s: Permission denied.\n"), Prog);
-				exit (E_NOPERM);
-			}
-			rflg++;
-			STRFCPY (roomno, optarg);
-			break;
 		case 'o':
 			if (!amroot) {
 				fprintf (stderr,
 				         _("%s: Permission denied.\n"), Prog);
 				exit (E_NOPERM);
 			}
-			oflg++;
+			oflg = true;
 			STRFCPY (slop, optarg);
 			break;
+		case 'r':
+			if (!may_change_field ('r')) {
+				fprintf (stderr,
+				         _("%s: Permission denied.\n"), Prog);
+				exit (E_NOPERM);
+			}
+			rflg = true;
+			STRFCPY (roomno, optarg);
+			break;
+		case 'R': /* no-op, handled in process_root_flag () */
+			break;
+		case 'u':
+			usage (E_SUCCESS);
+			/*@notreached@*/break;
 		case 'w':
 			if (!may_change_field ('w')) {
 				fprintf (stderr,
 				         _("%s: Permission denied.\n"), Prog);
 				exit (E_NOPERM);
 			}
-			wflg++;
+			wflg = true;
 			STRFCPY (workph, optarg);
 			break;
 		default:
-			usage ();
+			usage (E_USAGE);
 		}
 	}
 }
@@ -314,6 +358,7 @@ static void process_flags (int argc, char **argv)
 static void check_perms (const struct passwd *pw)
 {
 #ifdef USE_PAM
+	pam_handle_t *pamh = NULL;
 	int retval;
 	struct passwd *pampw;
 #endif
@@ -353,35 +398,34 @@ static void check_perms (const struct passwd *pw)
 	}
 
 #else				/* !USE_PAM */
-	retval = PAM_SUCCESS;
-
 	pampw = getpwuid (getuid ()); /* local, no need for xgetpwuid */
-	if (pampw == NULL) {
-		retval = PAM_USER_UNKNOWN;
-	}
-
-	if (retval == PAM_SUCCESS) {
-		retval = pam_start ("chfn", pampw->pw_name, &conv, &pamh);
-	}
-
-	if (retval == PAM_SUCCESS) {
-		retval = pam_authenticate (pamh, 0);
-		if (retval != PAM_SUCCESS) {
-			pam_end (pamh, retval);
-		}
-	}
-
-	if (retval == PAM_SUCCESS) {
-		retval = pam_acct_mgmt (pamh, 0);
-		if (retval != PAM_SUCCESS) {
-			pam_end (pamh, retval);
-		}
-	}
-
-	if (retval != PAM_SUCCESS) {
-		fprintf (stderr, _("%s: PAM authentication failed\n"), Prog);
+	if (NULL == pampw) {
+		fprintf (stderr,
+		         _("%s: Cannot determine your user name.\n"),
+		         Prog);
 		exit (E_NOPERM);
 	}
+
+	retval = pam_start ("chfn", pampw->pw_name, &conv, &pamh);
+
+	if (PAM_SUCCESS == retval) {
+		retval = pam_authenticate (pamh, 0);
+	}
+
+	if (PAM_SUCCESS == retval) {
+		retval = pam_acct_mgmt (pamh, 0);
+	}
+
+	if (PAM_SUCCESS != retval) {
+		fprintf (stderr, _("%s: PAM: %s\n"),
+		         Prog, pam_strerror (pamh, retval));
+		SYSLOG((LOG_ERR, "%s", pam_strerror (pamh, retval)));
+		if (NULL != pamh) {
+			(void) pam_end (pamh, retval);
+		}
+		exit (E_NOPERM);
+	}
+	(void) pam_end (pamh, retval);
 #endif				/* USE_PAM */
 }
 
@@ -404,8 +448,7 @@ static void update_gecos (const char *user, char *gecos)
 	if (setuid (0) != 0) {
 		fputs (_("Cannot change ID to root.\n"), stderr);
 		SYSLOG ((LOG_ERR, "can't setuid(0)"));
-		closelog ();
-		exit (E_NOPERM);
+		fail_exit (E_NOPERM);
 	}
 	pwd_init ();
 
@@ -414,18 +457,16 @@ static void update_gecos (const char *user, char *gecos)
 	 * password file. Get a lock on the file and open it.
 	 */
 	if (pw_lock () == 0) {
-		fputs (_("Cannot lock the password file; try again later.\n"),
-		       stderr);
-		SYSLOG ((LOG_WARN, "can't lock /etc/passwd"));
-		closelog ();
-		exit (E_NOPERM);
+		fprintf (stderr,
+		         _("%s: cannot lock %s; try again later.\n"),
+		         Prog, pw_dbname ());
+		fail_exit (E_NOPERM);
 	}
+	pw_locked = true;
 	if (pw_open (O_RDWR) == 0) {
-		fputs (_("Cannot open the password file.\n"), stderr);
-		pw_unlock ();
-		SYSLOG ((LOG_ERR, "can't open /etc/passwd"));
-		closelog ();
-		exit (E_NOPERM);
+		fprintf (stderr,
+		         _("%s: cannot open %s\n"), Prog, pw_dbname ());
+		fail_exit (E_NOPERM);
 	}
 
 	/*
@@ -436,10 +477,10 @@ static void update_gecos (const char *user, char *gecos)
 	 */
 	pw = pw_locate (user);
 	if (NULL == pw) {
-		pw_unlock ();
 		fprintf (stderr,
-		         _("%s: %s not found in /etc/passwd\n"), Prog, user);
-		exit (E_NOPERM);
+		         _("%s: user '%s' does not exist in %s\n"),
+		         Prog, user, pw_dbname ());
+		fail_exit (E_NOPERM);
 	}
 
 	/*
@@ -454,29 +495,26 @@ static void update_gecos (const char *user, char *gecos)
 	 * entry as well.
 	 */
 	if (pw_update (&pwent) == 0) {
-		fputs (_("Error updating the password entry.\n"), stderr);
-		pw_unlock ();
-		SYSLOG ((LOG_ERR, "error updating passwd entry"));
-		closelog ();
-		exit (E_NOPERM);
+		fprintf (stderr,
+		         _("%s: failed to prepare the new %s entry '%s'\n"),
+		         Prog, pw_dbname (), pwent.pw_name);
+		fail_exit (E_NOPERM);
 	}
 
 	/*
 	 * Changes have all been made, so commit them and unlock the file.
 	 */
 	if (pw_close () == 0) {
-		fputs (_("Cannot commit password file changes.\n"), stderr);
-		pw_unlock ();
-		SYSLOG ((LOG_ERR, "can't rewrite /etc/passwd"));
-		closelog ();
-		exit (E_NOPERM);
+		fprintf (stderr, _("%s: failure while writing changes to %s\n"), Prog, pw_dbname ());
+		SYSLOG ((LOG_ERR, "failure while writing changes to %s", pw_dbname ()));
+		fail_exit (E_NOPERM);
 	}
 	if (pw_unlock () == 0) {
-		fputs (_("Cannot unlock the password file.\n"), stderr);
-		SYSLOG ((LOG_ERR, "can't unlock /etc/passwd"));
-		closelog ();
-		exit (E_NOPERM);
+		fprintf (stderr, _("%s: failed to unlock %s\n"), Prog, pw_dbname ());
+		SYSLOG ((LOG_ERR, "failed to unlock %s", pw_dbname ()));
+		/* continue */
 	}
+	pw_locked = false;
 }
 
 /*
@@ -536,35 +574,40 @@ static void get_old_fields (const char *gecos)
  */
 static void check_fields (void)
 {
-	if (valid_field (fullnm, ":,=")) {
+	int err;
+	err = valid_field (fullnm, ":,=\n");
+	if (err > 0) {
+		fprintf (stderr, _("%s: name with non-ASCII characters: '%s'\n"), Prog, fullnm);
+	} else if (err < 0) {
 		fprintf (stderr, _("%s: invalid name: '%s'\n"), Prog, fullnm);
-		closelog ();
-		exit (E_NOPERM);
+		fail_exit (E_NOPERM);
 	}
-	if (valid_field (roomno, ":,=")) {
+	err = valid_field (roomno, ":,=\n");
+	if (err > 0) {
+		fprintf (stderr, _("%s: room number with non-ASCII characters: '%s'\n"), Prog, roomno);
+	} else if (err < 0) {
 		fprintf (stderr, _("%s: invalid room number: '%s'\n"),
 		         Prog, roomno);
-		closelog ();
-		exit (E_NOPERM);
+		fail_exit (E_NOPERM);
 	}
-	if (valid_field (workph, ":,=")) {
+	if (valid_field (workph, ":,=\n") != 0) {
 		fprintf (stderr, _("%s: invalid work phone: '%s'\n"),
 		         Prog, workph);
-		closelog ();
-		exit (E_NOPERM);
+		fail_exit (E_NOPERM);
 	}
-	if (valid_field (homeph, ":,=")) {
+	if (valid_field (homeph, ":,=\n") != 0) {
 		fprintf (stderr, _("%s: invalid home phone: '%s'\n"),
 		         Prog, homeph);
-		closelog ();
-		exit (E_NOPERM);
+		fail_exit (E_NOPERM);
 	}
-	if (valid_field (slop, ":")) {
+	err = valid_field (slop, ":\n");
+	if (err > 0) {
+		fprintf (stderr, _("%s: '%s' contains non-ASCII characters\n"), Prog, slop);
+	} else if (err < 0) {
 		fprintf (stderr,
 		         _("%s: '%s' contains illegal characters\n"),
 		         Prog, slop);
-		closelog ();
-		exit (E_NOPERM);
+		fail_exit (E_NOPERM);
 	}
 }
 
@@ -590,10 +633,18 @@ int main (int argc, char **argv)
 	char new_gecos[BUFSIZ];	/* buffer for new GECOS fields       */
 	char *user;
 
+	/*
+	 * Get the program name. The program name is used as a
+	 * prefix to most error messages.
+	 */
+	Prog = Basename (argv[0]);
+
 	sanitize_env ();
-	setlocale (LC_ALL, "");
-	bindtextdomain (PACKAGE, LOCALEDIR);
-	textdomain (PACKAGE);
+	(void) setlocale (LC_ALL, "");
+	(void) bindtextdomain (PACKAGE, LOCALEDIR);
+	(void) textdomain (PACKAGE);
+
+	process_root_flag ("-R", argc, argv);
 
 	/*
 	 * This command behaves different for root and non-root
@@ -601,12 +652,6 @@ int main (int argc, char **argv)
 	 */
 	amroot = (getuid () == 0);
 
-	/*
-	 * Get the program name. The program name is used as a
-	 * prefix to most error messages.
-	 */
-	Prog = Basename (argv[0]);
-
 	OPENLOG ("chfn");
 
 	/* parse the command line options */
@@ -620,18 +665,19 @@ int main (int argc, char **argv)
 		user = argv[optind];
 		pw = xgetpwnam (user);
 		if (NULL == pw) {
-			fprintf (stderr, _("%s: unknown user %s\n"), Prog,
+			fprintf (stderr, _("%s: user '%s' does not exist\n"), Prog,
 			         user);
-			exit (E_NOPERM);
+			fail_exit (E_NOPERM);
 		}
 	} else {
 		pw = get_my_pwent ();
 		if (NULL == pw) {
 			fprintf (stderr,
-			         _
-			         ("%s: Cannot determine your user name.\n"),
+			         _("%s: Cannot determine your user name.\n"),
 			         Prog);
-			exit (E_NOPERM);
+			SYSLOG ((LOG_WARN, "Cannot determine the user name of the caller (UID %lu)",
+			         (unsigned long) getuid ()));
+			fail_exit (E_NOPERM);
 		}
 		user = xstrdup (pw->pw_name);
 	}
@@ -655,7 +701,7 @@ int main (int argc, char **argv)
 			         ("%s: '%s' is the NIS master for this client.\n"),
 			         Prog, nis_master);
 		}
-		exit (E_NOPERM);
+		fail_exit (E_NOPERM);
 	}
 #endif
 
@@ -688,23 +734,19 @@ int main (int argc, char **argv)
 	if ((strlen (fullnm) + strlen (roomno) + strlen (workph) +
 	     strlen (homeph) + strlen (slop)) > (unsigned int) 80) {
 		fprintf (stderr, _("%s: fields too long\n"), Prog);
-		closelog ();
-		exit (E_NOPERM);
+		fail_exit (E_NOPERM);
 	}
 	snprintf (new_gecos, sizeof new_gecos, "%s,%s,%s,%s%s%s",
-		  fullnm, roomno, workph, homeph, slop[0] ? "," : "", slop);
+	          fullnm, roomno, workph, homeph,
+	          ('\0' != slop[0]) ? "," : "", slop);
 
 	/* Rewrite the user's gecos in the passwd file */
 	update_gecos (user, new_gecos);
 
-	SYSLOG ((LOG_INFO, "changed user `%s' information", user));
+	SYSLOG ((LOG_INFO, "changed user '%s' information", user));
 
 	nscd_flush_cache ("passwd");
 
-#ifdef USE_PAM
-	pam_end (pamh, PAM_SUCCESS);
-#endif				/* USE_PAM */
-
 	closelog ();
 	exit (E_SUCCESS);
 }
