@@ -1,5 +1,9 @@
 /*
- * Copyright 1992 - 1994, Julianne Frances Haugh
+ * Copyright (c) 1992 - 1994, Julianne Frances Haugh
+ * Copyright (c) 1996 - 2000, Marek Michałkiewicz
+ * Copyright (c) 2001       , Michał Moskal
+ * Copyright (c) 2001 - 2006, Tomasz Kłoczko
+ * Copyright (c) 2007 - 2011, Nicolas François
  * All rights reserved.
  *
  * Redistribution and use in source and binary forms, with or without
@@ -10,31 +14,32 @@
  * 2. Redistributions in binary form must reproduce the above copyright
  *    notice, this list of conditions and the following disclaimer in the
  *    documentation and/or other materials provided with the distribution.
- * 3. Neither the name of Julianne F. Haugh nor the names of its contributors
- *    may be used to endorse or promote products derived from this software
- *    without specific prior written permission.
+ * 3. The name of the copyright holders or contributors may not be used to
+ *    endorse or promote products derived from this software without
+ *    specific prior written permission.
  *
- * THIS SOFTWARE IS PROVIDED BY JULIE HAUGH AND CONTRIBUTORS ``AS IS'' AND
- * ANY EXPRESS OR IMPLIED WARRANTIES, INCLUDING, BUT NOT LIMITED TO, THE
- * IMPLIED WARRANTIES OF MERCHANTABILITY AND FITNESS FOR A PARTICULAR PURPOSE
- * ARE DISCLAIMED. IN NO EVENT SHALL JULIE HAUGH OR CONTRIBUTORS BE LIABLE
- * FOR ANY DIRECT, INDIRECT, INCIDENTAL, SPECIAL, EXEMPLARY, OR CONSEQUENTIAL
- * DAMAGES (INCLUDING, BUT NOT LIMITED TO, PROCUREMENT OF SUBSTITUTE GOODS
- * OR SERVICES; LOSS OF USE, DATA, OR PROFITS; OR BUSINESS INTERRUPTION)
- * HOWEVER CAUSED AND ON ANY THEORY OF LIABILITY, WHETHER IN CONTRACT, STRICT
- * LIABILITY, OR TORT (INCLUDING NEGLIGENCE OR OTHERWISE) ARISING IN ANY WAY
- * OUT OF THE USE OF THIS SOFTWARE, EVEN IF ADVISED OF THE POSSIBILITY OF
- * SUCH DAMAGE.
+ * THIS SOFTWARE IS PROVIDED BY THE COPYRIGHT HOLDERS AND CONTRIBUTORS
+ * ``AS IS'' AND ANY EXPRESS OR IMPLIED WARRANTIES, INCLUDING, BUT NOT
+ * LIMITED TO, THE IMPLIED WARRANTIES OF MERCHANTABILITY AND FITNESS FOR A
+ * PARTICULAR PURPOSE ARE DISCLAIMED.  IN NO EVENT SHALL THE COPYRIGHT
+ * HOLDERS OR CONTRIBUTORS BE LIABLE FOR ANY DIRECT, INDIRECT, INCIDENTAL,
+ * SPECIAL, EXEMPLARY, OR CONSEQUENTIAL DAMAGES (INCLUDING, BUT NOT
+ * LIMITED TO, PROCUREMENT OF SUBSTITUTE GOODS OR SERVICES; LOSS OF USE,
+ * DATA, OR PROFITS; OR BUSINESS INTERRUPTION) HOWEVER CAUSED AND ON ANY
+ * THEORY OF LIABILITY, WHETHER IN CONTRACT, STRICT LIABILITY, OR TORT
+ * (INCLUDING NEGLIGENCE OR OTHERWISE) ARISING IN ANY WAY OUT OF THE USE
+ * OF THIS SOFTWARE, EVEN IF ADVISED OF THE POSSIBILITY OF SUCH DAMAGE.
  */
 
 #include <config.h>
 
-#ident "$Id: grpck.c 1900 2008-03-17 23:00:49Z nekral-guest $"
+#ident "$Id: grpck.c 3559 2011-11-06 18:39:53Z nekral-guest $"
 
 #include <fcntl.h>
 #include <grp.h>
 #include <pwd.h>
 #include <stdio.h>
+#include <getopt.h>
 #include "chkname.h"
 #include "commonio.h"
 #include "defines.h"
@@ -49,8 +54,9 @@
 /*
  * Exit codes
  */
-
+/*@-exitarg@*/
 #define	E_OKAY		0
+#define	E_SUCCESS	0
 #define	E_USAGE		1
 #define	E_BAD_ENTRY	2
 #define	E_CANT_OPEN	3
@@ -60,71 +66,120 @@
 /*
  * Global variables
  */
-static char *Prog;
+const char *Prog;
+
 static const char *grp_file = GROUP_FILE;
-static int use_system_grp_file = 1;
+static bool use_system_grp_file = true;
 
 #ifdef	SHADOWGRP
 static const char *sgr_file = SGROUP_FILE;
-static int use_system_sgr_file = 1;
-static int is_shadow = 0;
+static bool use_system_sgr_file = true;
+static bool is_shadow = false;
+static bool sgr_locked = false;
 #endif
+static bool gr_locked = false;
 /* Options */
-static int read_only = 0;
-static int sort_mode = 0;
+static bool read_only = false;
+static bool sort_mode = false;
 
 /* local function prototypes */
-static void usage (void);
+static void fail_exit (int status);
+static /*@noreturn@*/void usage (int status);
 static void delete_member (char **, const char *);
 static void process_flags (int argc, char **argv);
 static void open_files (void);
-static void close_files (int changed);
+static void close_files (bool changed);
 static int check_members (const char *groupname,
                           char **members,
                           const char *fmt_info,
                           const char *fmt_prompt,
                           const char *fmt_syslog,
                           int *errors);
-static void check_grp_file (int *errors, int *changed);
+static void check_grp_file (int *errors, bool *changed);
 #ifdef SHADOWGRP
 static void compare_members_lists (const char *groupname,
                                    char **members,
                                    char **other_members,
                                    const char *file,
                                    const char *other_file);
-static void check_sgr_file (int *errors, int *changed);
+static void check_sgr_file (int *errors, bool *changed);
 #endif
 
+/*
+ * fail_exit - exit with an error code after unlocking files
+ */
+static void fail_exit (int status)
+{
+	if (gr_locked) {
+		if (gr_unlock () == 0) {
+			fprintf (stderr, _("%s: failed to unlock %s\n"), Prog, gr_dbname ());
+			SYSLOG ((LOG_ERR, "failed to unlock %s", gr_dbname ()));
+			/* continue */
+		}
+	}
+
+#ifdef	SHADOWGRP
+	if (sgr_locked) {
+		if (sgr_unlock () == 0) {
+			fprintf (stderr, _("%s: failed to unlock %s\n"), Prog, sgr_dbname ());
+			SYSLOG ((LOG_ERR, "failed to unlock %s", sgr_dbname ()));
+			/* continue */
+		}
+	}
+#endif
+
+	closelog ();
+
+	exit (status);
+}
+
 /*
  * usage - print syntax message and exit
  */
-static void usage (void)
+static /*@noreturn@*/void usage (int status)
 {
+	FILE *usageout = (E_SUCCESS != status) ? stderr : stdout;
 #ifdef	SHADOWGRP
-	fprintf (stderr, _("Usage: %s [-r] [-s] [group [gshadow]]\n"), Prog);
-#else
-	fprintf (stderr, _("Usage: %s [-r] [-s] [group]\n"), Prog);
-#endif
-	exit (E_USAGE);
+	(void) fprintf (usageout,
+	                _("Usage: %s [options] [group [gshadow]]\n"
+	                  "\n"
+	                  "Options:\n"),
+	                Prog);
+#else				/* !SHADOWGRP */
+	(void) fprintf (usageout,
+	                _("Usage: %s [options] [group]\n"
+	                  "\n"
+	                  "Options:\n"),
+	                Prog);
+#endif				/* !SHADOWGRP */
+	(void) fputs (_("  -h, --help                    display this help message and exit\n"), usageout);
+	(void) fputs (_("  -r, --read-only               display errors and warnings\n"
+	                "                                but do not change files\n"), usageout);
+	(void) fputs (_("  -R, --root CHROOT_DIR         directory to chroot into\n"), usageout);
+	(void) fputs (_("  -s, --sort                    sort entries by UID\n"), usageout);
+	(void) fputs ("\n", usageout);
+	exit (status);
 }
 
 /*
  * delete_member - delete an entry in a list of members
+ *
+ * It only deletes the first entry with the given name.
+ * The member is defined by its address, no string comparison are
+ * performed.
  */
 static void delete_member (char **list, const char *member)
 {
 	int i;
 
-	for (i = 0; list[i]; i++) {
+	for (i = 0; NULL != list[i]; i++) {
 		if (list[i] == member) {
 			break;
 		}
 	}
 
-	if (list[i]) {
-		for (; list[i]; i++) {
-			list[i] = list[i + 1];
-		}
+	for (; NULL != list[i]; i++) {
+		list[i] = list[i + 1];
 	}
 }
 
@@ -135,29 +190,43 @@ static void delete_member (char **list, const char *member)
  */
 static void process_flags (int argc, char **argv)
 {
-	int arg;
+	int c;
+	static struct option long_options[] = {
+		{"help",      no_argument,       NULL, 'h'},
+		{"quiet",     no_argument,       NULL, 'q'},
+		{"read-only", no_argument,       NULL, 'r'},
+		{"root",      required_argument, NULL, 'R'},
+		{"sort",      no_argument,       NULL, 's'},
+		{NULL, 0, NULL, '\0'}
+	};
 
 	/*
 	 * Parse the command line arguments
 	 */
-	while ((arg = getopt (argc, argv, "qrs")) != EOF) {
-		switch (arg) {
+	while ((c = getopt_long (argc, argv, "hqrR:s",
+	                         long_options, NULL)) != -1) {
+		switch (c) {
+		case 'h':
+			usage (E_SUCCESS);
+			/*@notreached@*/break;
 		case 'q':
 			/* quiet - ignored for now */
 			break;
 		case 'r':
-			read_only = 1;
+			read_only = true;
+			break;
+		case 'R': /* no-op, handled in process_root_flag () */
 			break;
 		case 's':
-			sort_mode = 1;
+			sort_mode = true;
 			break;
 		default:
-			usage ();
+			usage (E_USAGE);
 		}
 	}
 
 	if (sort_mode && read_only) {
-		fprintf (stderr, _("%s: -s and -r are incompatibile\n"), Prog);
+		fprintf (stderr, _("%s: -s and -r are incompatible\n"), Prog);
 		exit (E_USAGE);
 	}
 
@@ -165,12 +234,12 @@ static void process_flags (int argc, char **argv)
 	 * Make certain we have the right number of arguments
 	 */
 #ifdef	SHADOWGRP
-	if ((argc < optind) || (argc > (optind + 2)))
+	if (argc > (optind + 2))
 #else
-	if ((argc < optind) || (argc > (optind + 1)))
+	if (argc > (optind + 1))
 #endif
 	{
-		usage ();
+		usage (E_USAGE);
 	}
 
 	/*
@@ -179,15 +248,15 @@ static void process_flags (int argc, char **argv)
 	 */
 	if (optind != argc) {
 		grp_file = argv[optind];
-		gr_name (grp_file);
-		use_system_grp_file = 0;
+		gr_setdbname (grp_file);
+		use_system_grp_file = false;
 	}
 #ifdef	SHADOWGRP
 	if ((optind + 2) == argc) {
 		sgr_file = argv[optind + 1];
-		sgr_name (sgr_file);
-		is_shadow = 1;
-		use_system_sgr_file = 0;
+		sgr_setdbname (sgr_file);
+		is_shadow = true;
+		use_system_sgr_file = false;
 	} else if (optind == argc) {
 		is_shadow = sgr_file_present ();
 	}
@@ -207,23 +276,21 @@ static void open_files (void)
 	 */
 	if (!read_only) {
 		if (gr_lock () == 0) {
-			fprintf (stderr, _("%s: cannot lock file %s\n"),
+			fprintf (stderr,
+			         _("%s: cannot lock %s; try again later.\n"),
 			         Prog, grp_file);
-			if (use_system_grp_file) {
-				SYSLOG ((LOG_WARN, "cannot lock %s", grp_file));
-			}
-			closelog ();
-			exit (E_CANT_LOCK);
+			fail_exit (E_CANT_LOCK);
 		}
+		gr_locked = true;
 #ifdef	SHADOWGRP
-		if (is_shadow && (sgr_lock () == 0)) {
-			fprintf (stderr, _("%s: cannot lock file %s\n"),
-			         Prog, sgr_file);
-			if (use_system_sgr_file) {
-				SYSLOG ((LOG_WARN, "cannot lock %s", sgr_file));
+		if (is_shadow) {
+			if (sgr_lock () == 0) {
+				fprintf (stderr,
+				         _("%s: cannot lock %s; try again later.\n"),
+				         Prog, sgr_file);
+				fail_exit (E_CANT_LOCK);
 			}
-			closelog ();
-			exit (E_CANT_LOCK);
+			sgr_locked = true;
 		}
 #endif
 	}
@@ -233,23 +300,21 @@ static void open_files (void)
 	 * O_RDWR otherwise.
 	 */
 	if (gr_open (read_only ? O_RDONLY : O_RDWR) == 0) {
-		fprintf (stderr, _("%s: cannot open file %s\n"), Prog,
+		fprintf (stderr, _("%s: cannot open %s\n"), Prog,
 		         grp_file);
 		if (use_system_grp_file) {
 			SYSLOG ((LOG_WARN, "cannot open %s", grp_file));
 		}
-		closelog ();
-		exit (E_CANT_OPEN);
+		fail_exit (E_CANT_OPEN);
 	}
 #ifdef	SHADOWGRP
 	if (is_shadow && (sgr_open (read_only ? O_RDONLY : O_RDWR) == 0)) {
-		fprintf (stderr, _("%s: cannot open file %s\n"), Prog,
+		fprintf (stderr, _("%s: cannot open %s\n"), Prog,
 		         sgr_file);
 		if (use_system_sgr_file) {
 			SYSLOG ((LOG_WARN, "cannot open %s", sgr_file));
 		}
-		closelog ();
-		exit (E_CANT_OPEN);
+		fail_exit (E_CANT_OPEN);
 	}
 #endif
 }
@@ -261,7 +326,7 @@ static void open_files (void)
  *	changes are committed in the databases. The databases are
  *	unlocked anyway.
  */
-static void close_files (int changed)
+static void close_files (bool changed)
 {
 	/*
 	 * All done. If there were no change we can just abandon any
@@ -269,15 +334,15 @@ static void close_files (int changed)
 	 */
 	if (changed) {
 		if (gr_close () == 0) {
-			fprintf (stderr, _("%s: cannot update file %s\n"),
+			fprintf (stderr, _("%s: failure while writing changes to %s\n"),
 			         Prog, grp_file);
-			exit (E_CANT_UPDATE);
+			fail_exit (E_CANT_UPDATE);
 		}
 #ifdef	SHADOWGRP
 		if (is_shadow && (sgr_close () == 0)) {
-			fprintf (stderr, _("%s: cannot update file %s\n"),
+			fprintf (stderr, _("%s: failure while writing changes to %s\n"),
 			         Prog, sgr_file);
-			exit (E_CANT_UPDATE);
+			fail_exit (E_CANT_UPDATE);
 		}
 #endif
 	}
@@ -286,11 +351,23 @@ static void close_files (int changed)
 	 * Don't be anti-social - unlock the files when you're done.
 	 */
 #ifdef	SHADOWGRP
-	if (is_shadow) {
-		sgr_unlock ();
+	if (sgr_locked) {
+		if (sgr_unlock () == 0) {
+			fprintf (stderr, _("%s: failed to unlock %s\n"), Prog, sgr_dbname ());
+			SYSLOG ((LOG_ERR, "failed to unlock %s", sgr_dbname ()));
+			/* continue */
+		}
+		sgr_locked = false;
 	}
 #endif
-	(void) gr_unlock ();
+	if (gr_locked) {
+		if (gr_unlock () == 0) {
+			fprintf (stderr, _("%s: failed to unlock %s\n"), Prog, gr_dbname ());
+			SYSLOG ((LOG_ERR, "failed to unlock %s", gr_dbname ()));
+			/* continue */
+		}
+		gr_locked = false;
+	}
 }
 
 /*
@@ -345,6 +422,9 @@ static int check_members (const char *groupname,
 		SYSLOG ((LOG_INFO, fmt_syslog, members[i], groupname));
 		members_changed = 1;
 		delete_member (members, members[i]);
+
+		/* Rewind in case of removal */
+		i--;
 	}
 
 	return members_changed;
@@ -387,7 +467,7 @@ static void compare_members_lists (const char *groupname,
 /*
  * check_grp_file - check the content of the group file
  */
-static void check_grp_file (int *errors, int *changed)
+static void check_grp_file (int *errors, bool *changed)
 {
 	struct commonio_entry *gre, *tgre;
 	struct group *grp;
@@ -418,7 +498,7 @@ static void check_grp_file (int *errors, int *changed)
 			 * Tell the user this entire line is bogus and ask
 			 * them to delete it.
 			 */
-			puts (_("invalid group file entry"));
+			(void) puts (_("invalid group file entry"));
 			printf (_("delete line '%s'? "), gre->line);
 			*errors += 1;
 
@@ -436,9 +516,9 @@ static void check_grp_file (int *errors, int *changed)
 			 * to try out the next list element.
 			 */
 		      delete_gr:
-			SYSLOG ((LOG_INFO, "delete group line `%s'",
+			SYSLOG ((LOG_INFO, "delete group line '%s'",
 			         gre->line));
-			*changed = 1;
+			*changed = true;
 
 			__gr_del_entry (gre);
 			continue;
@@ -478,7 +558,7 @@ static void check_grp_file (int *errors, int *changed)
 			 * Tell the user this entry is a duplicate of
 			 * another and ask them to delete it.
 			 */
-			puts (_("duplicate group entry"));
+			(void) puts (_("duplicate group entry"));
 			printf (_("delete line '%s'? "), gre->line);
 			*errors += 1;
 
@@ -493,11 +573,19 @@ static void check_grp_file (int *errors, int *changed)
 		/*
 		 * Check for invalid group names.  --marekm
 		 */
-		if (!check_group_name (grp->gr_name)) {
+		if (!is_valid_group_name (grp->gr_name)) {
 			*errors += 1;
 			printf (_("invalid group name '%s'\n"), grp->gr_name);
 		}
 
+		/*
+		 * Check for invalid group ID.
+		 */
+		if (grp->gr_gid == (gid_t)-1) {
+			printf (_("invalid group ID '%lu'\n"), (long unsigned int)grp->gr_gid);
+			*errors += 1;
+		}
+
 		/*
 		 * Workaround for a NYS libc 5.3.12 bug on RedHat 4.2 -
 		 * groups with no members are returned as groups with one
@@ -512,10 +600,10 @@ static void check_grp_file (int *errors, int *changed)
 		if (check_members (grp->gr_name, grp->gr_mem,
 		                   _("group %s: no user %s\n"),
 		                   _("delete member '%s'? "),
-		                   "delete member `%s' from group `%s'",
+		                   "delete member '%s' from group '%s'",
 		                   errors) == 1) {
-			*changed = 1;
-			gre->changed = 1;
+			*changed = true;
+			gre->changed = true;
 			__gr_set_changed ();
 		}
 
@@ -527,10 +615,9 @@ static void check_grp_file (int *errors, int *changed)
 		if (is_shadow) {
 			sgr = (struct sgrp *) sgr_locate (grp->gr_name);
 			if (sgr == NULL) {
-				printf (_
-				        ("no matching group file entry in %s\n"),
+				printf (_("no matching group file entry in %s\n"),
 				        sgr_file);
-				printf (_("add group '%s' in %s ?"),
+				printf (_("add group '%s' in %s? "),
 				        grp->gr_name, sgr_file);
 				*errors += 1;
 				if (yes_or_no (read_only)) {
@@ -543,26 +630,24 @@ static void check_grp_file (int *errors, int *changed)
 					sg.sg_adm = &empty;
 					sg.sg_mem = grp->gr_mem;
 					SYSLOG ((LOG_INFO,
-					         "add group `%s' to `%s'",
+					         "add group '%s' to '%s'",
 					         grp->gr_name, sgr_file));
-					*changed = 1;
+					*changed = true;
 
 					if (sgr_update (&sg) == 0) {
 						fprintf (stderr,
-						         _
-						         ("%s: can't update shadow entry for %s\n"),
-						         Prog, sg.sg_name);
-						exit (E_CANT_UPDATE);
+						         _("%s: failed to prepare the new %s entry '%s'\n"),
+						         Prog, sgr_dbname (), sg.sg_name);
+						fail_exit (E_CANT_UPDATE);
 					}
 					/* remove password from /etc/group */
 					gr = *grp;
 					gr.gr_passwd = SHADOW_PASSWD_STRING;	/* XXX warning: const */
 					if (gr_update (&gr) == 0) {
 						fprintf (stderr,
-						         _
-						         ("%s: can't update entry for group %s\n"),
-						         Prog, gr.gr_name);
-						exit (E_CANT_UPDATE);
+						         _("%s: failed to prepare the new %s entry '%s'\n"),
+						         Prog, gr_dbname (), gr.gr_name);
+						fail_exit (E_CANT_UPDATE);
 					}
 				}
 			} else {
@@ -573,6 +658,15 @@ static void check_grp_file (int *errors, int *changed)
 				compare_members_lists (grp->gr_name,
 				                       grp->gr_mem, sgr->sg_mem,
 				                       grp_file, sgr_file);
+
+				/* The group entry has a gshadow counterpart.
+				 * Make sure no passwords are in group.
+				 */
+				if (strcmp (grp->gr_passwd, SHADOW_PASSWD_STRING) != 0) {
+					printf (_("group %s has an entry in %s, but its password field in %s is not set to 'x'\n"),
+					        grp->gr_name, sgr_file, grp_file);
+					*errors += 1;
+				}
 			}
 		}
 #endif
@@ -584,7 +678,7 @@ static void check_grp_file (int *errors, int *changed)
 /*
  * check_sgr_file - check the content of the shadowed group file (gshadow)
  */
-static void check_sgr_file (int *errors, int *changed)
+static void check_sgr_file (int *errors, bool *changed)
 {
 	struct group *grp;
 	struct commonio_entry *sge, *tsge;
@@ -606,7 +700,7 @@ static void check_sgr_file (int *errors, int *changed)
 			 * Tell the user this entire line is bogus and ask
 			 * them to delete it.
 			 */
-			puts (_("invalid shadow group file entry"));
+			(void) puts (_("invalid shadow group file entry"));
 			printf (_("delete line '%s'? "), sge->line);
 			*errors += 1;
 
@@ -624,9 +718,9 @@ static void check_sgr_file (int *errors, int *changed)
 			 * of the loop to try out the next list element.
 			 */
 		      delete_sg:
-			SYSLOG ((LOG_INFO, "delete shadow line `%s'",
+			SYSLOG ((LOG_INFO, "delete shadow line '%s'",
 			         sge->line));
-			*changed = 1;
+			*changed = true;
 
 			__sgr_del_entry (sge);
 			continue;
@@ -666,7 +760,7 @@ static void check_sgr_file (int *errors, int *changed)
 			 * Tell the user this entry is a duplicate of
 			 * another and ask them to delete it.
 			 */
-			puts (_("duplicate shadow group entry"));
+			(void) puts (_("duplicate shadow group entry"));
 			printf (_("delete line '%s'? "), sge->line);
 			*errors += 1;
 
@@ -706,10 +800,10 @@ static void check_sgr_file (int *errors, int *changed)
 		if (check_members (sgr->sg_name, sgr->sg_adm,
 		                   _("shadow group %s: no administrative user %s\n"),
 		                   _("delete administrative member '%s'? "),
-		                   "delete admin `%s' from shadow group `%s'",
+		                   "delete admin '%s' from shadow group '%s'",
 		                   errors) == 1) {
-			*changed = 1;
-			sge->changed = 1;
+			*changed = true;
+			sge->changed = true;
 			__sgr_set_changed ();
 		}
 
@@ -719,10 +813,10 @@ static void check_sgr_file (int *errors, int *changed)
 		if (check_members (sgr->sg_name, sgr->sg_mem,
 		                   _("shadow group %s: no user %s\n"),
 		                   _("delete member '%s'? "),
-		                   "delete member `%s' from shadow group `%s'",
+		                   "delete member '%s' from shadow group '%s'",
 		                   errors) == 1) {
-			*changed = 1;
-			sge->changed = 1;
+			*changed = true;
+			sge->changed = true;
 			__sgr_set_changed ();
 		}
 	}
@@ -735,16 +829,18 @@ static void check_sgr_file (int *errors, int *changed)
 int main (int argc, char **argv)
 {
 	int errors = 0;
-	int changed = 0;
+	bool changed = false;
 
 	/*
 	 * Get my name so that I can use it to report errors.
 	 */
 	Prog = Basename (argv[0]);
 
-	setlocale (LC_ALL, "");
-	bindtextdomain (PACKAGE, LOCALEDIR);
-	textdomain (PACKAGE);
+	(void) setlocale (LC_ALL, "");
+	(void) bindtextdomain (PACKAGE, LOCALEDIR);
+	(void) textdomain (PACKAGE);
+
+	process_root_flag ("-R", argc, argv);
 
 	OPENLOG ("grpck");
 
@@ -759,7 +855,7 @@ int main (int argc, char **argv)
 		if (is_shadow) {
 			sgr_sort ();
 		}
-		changed = 1;
+		changed = true;
 #endif
 	} else {
 		check_grp_file (&errors, &changed);
@@ -778,12 +874,14 @@ int main (int argc, char **argv)
 	/*
 	 * Tell the user what we did and exit.
 	 */
-	if (errors != 0) {
-		printf (changed ?
-		        _("%s: the files have been updated\n") :
-		        _("%s: no changes\n"), Prog);
+	if (0 != errors) {
+		if (changed) {
+			printf (_("%s: the files have been updated\n"), Prog);
+		} else {
+			printf (_("%s: no changes\n"), Prog);
+		}
 	}
 
-	exit (errors ? E_BAD_ENTRY : E_OKAY);
+	return ((0 != errors) ? E_BAD_ENTRY : E_OKAY);
 }
 
