@@ -2,7 +2,7 @@
  * Copyright (c) 1991 - 1994, Julianne Frances Haugh
  * Copyright (c) 1996 - 2000, Marek Michałkiewicz
  * Copyright (c) 2000 - 2006, Tomasz Kłoczko
- * Copyright (c) 2007 - 2009, Nicolas François
+ * Copyright (c) 2007 - 2012, Nicolas François
  * All rights reserved.
  *
  * Redistribution and use in source and binary forms, with or without
@@ -32,7 +32,7 @@
 
 #include <config.h>
 
-#ident "$Id: useradd.c 2901 2009-05-10 18:26:33Z nekral-guest $"
+#ident "$Id: useradd.c 3743 2012-05-25 11:51:53Z nekral-guest $"
 
 #include <assert.h>
 #include <ctype.h>
@@ -65,6 +65,9 @@
 #include "sgroupio.h"
 #endif
 #include "shadowio.h"
+#ifdef WITH_TCB
+#include "tcbfuncs.h"
+#endif
 
 #ifndef SKEL_DIR
 #define SKEL_DIR "/etc/skel"
@@ -82,7 +85,7 @@
 /*
  * Global variables
  */
-char *Prog;
+const char *Prog;
 
 /*
  * These defaults are used if there is no defaults file.
@@ -97,8 +100,6 @@ static const char *def_create_mail_spool = "no";
 static long def_inactive = -1;
 static const char *def_expire = "";
 
-static char def_file[] = USER_DEFAULTS_FILE;
-
 #define	VALID(s)	(strcspn (s, ":\n") == strlen (s))
 
 static const char *user_name = "";
@@ -110,8 +111,8 @@ static const char *user_home = "";
 static const char *user_shell = "";
 static const char *create_mail_spool = "";
 #ifdef WITH_SELINUX
-static const char *user_selinux = "";
-#endif
+static /*@notnull@*/const char *user_selinux = "";
+#endif				/* WITH_SELINUX */
 
 static long user_expire = -1;
 static bool is_shadow_pwd;
@@ -145,8 +146,11 @@ static bool
     rflg = false,		/* create a system account */
     sflg = false,		/* shell program for new account */
     uflg = false,		/* specify user ID for new account */
-    Uflg = false,		/* create a group having the same name as the user */
-    Zflg = false;		/* new selinux user */
+    Uflg = false;		/* create a group having the same name as the user */
+
+#ifdef WITH_SELINUX
+#define Zflg ('\0' != *user_selinux)
+#endif				/* WITH_SELINUX */
 
 static bool home_added = false;
 
@@ -163,15 +167,15 @@ static bool home_added = false;
 #define E_NAME_IN_USE	9	/* username already in use */
 #define E_GRP_UPDATE	10	/* can't update group file */
 #define E_HOMEDIR	12	/* can't create home directory */
-#define	E_MAIL_SPOOL	13	/* can't create mail spool */
+#define E_SE_UPDATE	14	/* can't update SELinux user mapping */
 
 #define DGROUP			"GROUP="
-#define HOME			"HOME="
-#define SHELL			"SHELL="
-#define INACT			"INACTIVE="
-#define EXPIRE			"EXPIRE="
-#define SKEL			"SKEL="
-#define CREATE_MAIL_SPOOL	"CREATE_MAIL_SPOOL="
+#define DHOME			"HOME="
+#define DSHELL			"SHELL="
+#define DINACT			"INACTIVE="
+#define DEXPIRE			"EXPIRE="
+#define DSKEL			"SKEL="
+#define DCREATE_MAIL_SPOOL	"CREATE_MAIL_SPOOL="
 
 /* local function prototypes */
 static void fail_exit (int);
@@ -179,11 +183,8 @@ static void get_defaults (void);
 static void show_defaults (void);
 static int set_defaults (void);
 static int get_groups (char *);
-static void usage (void);
+static void usage (int status);
 static void new_pwent (struct passwd *);
-#ifdef WITH_SELINUX
-static void selinux_update_mapping (void);
-#endif
 
 static long scale_age (long);
 static void new_spent (struct spwd *);
@@ -192,6 +193,7 @@ static void grp_update (void);
 static void process_flags (int argc, char **argv);
 static void close_files (void);
 static void open_files (void);
+static void open_shadow (void);
 static void faillog_reset (uid_t);
 static void lastlog_reset (uid_t);
 static void usr_update (void);
@@ -204,7 +206,12 @@ static void create_mail (void);
 static void fail_exit (int code)
 {
 	if (home_added) {
-		rmdir (user_home);
+		if (rmdir (user_home) != 0) {
+			fprintf (stderr,
+			         _("%s: %s was created, but could not be removed\n"),
+			         Prog, user_home);
+			SYSLOG ((LOG_ERR, "failed to remove %s", user_home));
+		}
 	}
 
 	if (spw_locked) {
@@ -291,7 +298,7 @@ static void get_defaults (void)
 	 * Open the defaults file for reading.
 	 */
 
-	fp = fopen (def_file, "r");
+	fp = fopen (USER_DEFAULTS_FILE, "r");
 	if (NULL == fp) {
 		return;
 	}
@@ -324,7 +331,7 @@ static void get_defaults (void)
 				         Prog, cp);
 				fprintf (stderr,
 				         _("%s: the %s configuration in %s will be ignored\n"),
-				         Prog, DGROUP, def_file);
+				         Prog, DGROUP, USER_DEFAULTS_FILE);
 			} else {
 				def_group = grp->gr_gid;
 				def_gname = xstrdup (grp->gr_name);
@@ -334,29 +341,29 @@ static void get_defaults (void)
 		/*
 		 * Default HOME filesystem
 		 */
-		else if (MATCH (buf, HOME)) {
+		else if (MATCH (buf, DHOME)) {
 			def_home = xstrdup (cp);
 		}
 
 		/*
 		 * Default Login Shell command
 		 */
-		else if (MATCH (buf, SHELL)) {
+		else if (MATCH (buf, DSHELL)) {
 			def_shell = xstrdup (cp);
 		}
 
 		/*
 		 * Default Password Inactive value
 		 */
-		else if (MATCH (buf, INACT)) {
+		else if (MATCH (buf, DINACT)) {
 			if (   (getlong (cp, &def_inactive) == 0)
 			    || (def_inactive < -1)) {
 				fprintf (stderr,
 				         _("%s: invalid numeric argument '%s'\n"),
-				         Prog, optarg);
+				         Prog, cp);
 				fprintf (stderr,
 				         _("%s: the %s configuration in %s will be ignored\n"),
-				         Prog, INACT, def_file);
+				         Prog, DINACT, USER_DEFAULTS_FILE);
 				def_inactive = -1;
 			}
 		}
@@ -364,14 +371,14 @@ static void get_defaults (void)
 		/*
 		 * Default account expiration date
 		 */
-		else if (MATCH (buf, EXPIRE)) {
+		else if (MATCH (buf, DEXPIRE)) {
 			def_expire = xstrdup (cp);
 		}
 
 		/*
 		 * Default Skeleton information
 		 */
-		else if (MATCH (buf, SKEL)) {
+		else if (MATCH (buf, DSKEL)) {
 			if ('\0' == *cp) {
 				cp = SKEL_DIR;	/* XXX warning: const */
 			}
@@ -382,9 +389,9 @@ static void get_defaults (void)
 		/*
 		 * Create by default user mail spool or not ?
 		 */
-		else if (MATCH (buf, CREATE_MAIL_SPOOL)) {
+		else if (MATCH (buf, DCREATE_MAIL_SPOOL)) {
 			if (*cp == '\0') {
-				cp = CREATE_MAIL_SPOOL;	/* XXX warning: const */
+				cp = "no";	/* XXX warning: const */
 			}
 
 			def_create_mail_spool = xstrdup (cp);
@@ -458,7 +465,7 @@ static int set_defaults (void)
 	 * temporary file, using any new values. Each line is checked
 	 * to insure that it is not output more than once.
 	 */
-	ifp = fopen (def_file, "r");
+	ifp = fopen (USER_DEFAULTS_FILE, "r");
 	if (NULL == ifp) {
 		fprintf (ofp, "# useradd defaults file\n");
 		goto skip;
@@ -475,7 +482,8 @@ static int set_defaults (void)
 			if (feof (ifp) == 0) {
 				fprintf (stderr,
 				         _("%s: line too long in %s: %s..."),
-				         Prog, def_file, buf);
+				         Prog, USER_DEFAULTS_FILE, buf);
+				(void) fclose (ifp);
 				return -1;
 			}
 		}
@@ -483,25 +491,25 @@ static int set_defaults (void)
 		if (!out_group && MATCH (buf, DGROUP)) {
 			fprintf (ofp, DGROUP "%u\n", (unsigned int) def_group);
 			out_group = true;
-		} else if (!out_home && MATCH (buf, HOME)) {
-			fprintf (ofp, HOME "%s\n", def_home);
+		} else if (!out_home && MATCH (buf, DHOME)) {
+			fprintf (ofp, DHOME "%s\n", def_home);
 			out_home = true;
-		} else if (!out_inactive && MATCH (buf, INACT)) {
-			fprintf (ofp, INACT "%ld\n", def_inactive);
+		} else if (!out_inactive && MATCH (buf, DINACT)) {
+			fprintf (ofp, DINACT "%ld\n", def_inactive);
 			out_inactive = true;
-		} else if (!out_expire && MATCH (buf, EXPIRE)) {
-			fprintf (ofp, EXPIRE "%s\n", def_expire);
+		} else if (!out_expire && MATCH (buf, DEXPIRE)) {
+			fprintf (ofp, DEXPIRE "%s\n", def_expire);
 			out_expire = true;
-		} else if (!out_shell && MATCH (buf, SHELL)) {
-			fprintf (ofp, SHELL "%s\n", def_shell);
+		} else if (!out_shell && MATCH (buf, DSHELL)) {
+			fprintf (ofp, DSHELL "%s\n", def_shell);
 			out_shell = true;
-		} else if (!out_skel && MATCH (buf, SKEL)) {
-			fprintf (ofp, SKEL "%s\n", def_template);
+		} else if (!out_skel && MATCH (buf, DSKEL)) {
+			fprintf (ofp, DSKEL "%s\n", def_template);
 			out_skel = true;
 		} else if (!out_create_mail_spool
-			   && MATCH (buf, CREATE_MAIL_SPOOL)) {
+			   && MATCH (buf, DCREATE_MAIL_SPOOL)) {
 			fprintf (ofp,
-			         CREATE_MAIL_SPOOL "%s\n",
+			         DCREATE_MAIL_SPOOL "%s\n",
 			         def_create_mail_spool);
 			out_create_mail_spool = true;
 		} else
@@ -518,18 +526,18 @@ static int set_defaults (void)
 	if (!out_group)
 		fprintf (ofp, DGROUP "%u\n", (unsigned int) def_group);
 	if (!out_home)
-		fprintf (ofp, HOME "%s\n", def_home);
+		fprintf (ofp, DHOME "%s\n", def_home);
 	if (!out_inactive)
-		fprintf (ofp, INACT "%ld\n", def_inactive);
+		fprintf (ofp, DINACT "%ld\n", def_inactive);
 	if (!out_expire)
-		fprintf (ofp, EXPIRE "%s\n", def_expire);
+		fprintf (ofp, DEXPIRE "%s\n", def_expire);
 	if (!out_shell)
-		fprintf (ofp, SHELL "%s\n", def_shell);
+		fprintf (ofp, DSHELL "%s\n", def_shell);
 	if (!out_skel)
-		fprintf (ofp, SKEL "%s\n", def_template);
+		fprintf (ofp, DSKEL "%s\n", def_template);
 
 	if (!out_create_mail_spool)
-		fprintf (ofp, CREATE_MAIL_SPOOL "%s\n", def_create_mail_spool);
+		fprintf (ofp, DCREATE_MAIL_SPOOL "%s\n", def_create_mail_spool);
 
 	/*
 	 * Flush and close the file. Check for errors to make certain
@@ -546,13 +554,14 @@ static int set_defaults (void)
 	/*
 	 * Rename the current default file to its backup name.
 	 */
-	wlen = snprintf (buf, sizeof buf, "%s-", def_file);
+	wlen = snprintf (buf, sizeof buf, "%s-", USER_DEFAULTS_FILE);
 	assert (wlen < (int) sizeof buf);
-	if ((rename (def_file, buf) != 0) && (ENOENT != errno)) {
+	unlink (buf);
+	if ((link (USER_DEFAULTS_FILE, buf) != 0) && (ENOENT != errno)) {
 		int err = errno;
 		fprintf (stderr,
-		         _("%s: rename: %s: %s"),
-		         Prog, def_file, strerror (err));
+		         _("%s: Cannot create backup file (%s): %s\n"),
+		         Prog, buf, strerror (err));
 		unlink (new_file);
 		return -1;
 	}
@@ -560,10 +569,10 @@ static int set_defaults (void)
 	/*
 	 * Rename the new default file to its correct name.
 	 */
-	if (rename (new_file, def_file) != 0) {
+	if (rename (new_file, USER_DEFAULTS_FILE) != 0) {
 		int err = errno;
 		fprintf (stderr,
-		         _("%s: rename: %s: %s"),
+		         _("%s: rename: %s: %s\n"),
 		         Prog, new_file, strerror (err));
 		return -1;
 	}
@@ -623,6 +632,8 @@ static int get_groups (char *list)
 		/*
 		 * There must be a match, either by GID value or by
 		 * string name.
+		 * FIXME: It should exist according to gr_locate,
+		 *        otherwise, we can't change its members
 		 */
 		if (NULL == grp) {
 			fprintf (stderr,
@@ -681,52 +692,49 @@ static int get_groups (char *list)
 /*
  * usage - display usage message and exit
  */
-static void usage (void)
+static void usage (int status)
 {
-	fprintf (stderr,
-	         _("Usage: useradd [options] LOGIN\n"
-	         "\n"
-	         "Options:\n"
-	         "  -b, --base-dir BASE_DIR       base directory for the new user account\n"
-	         "                                home directory\n"
-	         "  -c, --comment COMMENT         set the GECOS field for the new user account\n"
-	         "  -d, --home-dir HOME_DIR       home directory for the new user account\n"
-	         "  -D, --defaults                print or save modified default useradd\n"
-	         "                                configuration\n"
-	         "  -e, --expiredate EXPIRE_DATE  set account expiration date to EXPIRE_DATE\n"
-	         "  -f, --inactive INACTIVE       set password inactive after expiration\n"
-	         "                                to INACTIVE\n"
-	         "  -g, --gid GROUP               force use GROUP for the new user account\n"
-	         "  -G, --groups GROUPS           list of supplementary groups for the new\n"
-	         "                                user account\n"
-	         "  -h, --help                    display this help message and exit\n"
-	         "  -k, --skel SKEL_DIR           specify an alternative skel directory\n"
-	         "  -K, --key KEY=VALUE           overrides /etc/login.defs defaults\n"
-	         "  -l, --no-log-init             do not add the user to the lastlog and\n"
-	         "                                faillog databases\n"
-	         "  -m, --create-home             create home directory for the new user\n"
-	         "                                account\n"
-	         "  -M, --no-create-home          do not create user's home directory\n"
-	         "                                (overrides /etc/login.defs)\n"
-	         "  -N, --no-user-group           do not create a group with the same name as\n"
-	         "                                the user\n"
-	         "  -o, --non-unique              allow create user with duplicate\n"
-	         "                                (non-unique) UID\n"
-	         "  -p, --password PASSWORD       use encrypted password for the new user\n"
-	         "                                account\n"
-	         "  -r, --system                  create a system account\n"
-	         "  -s, --shell SHELL             the login shell for the new user account\n"
-	         "  -u, --uid UID                 force use the UID for the new user account\n"
-	         "  -U, --user-group              create a group with the same name as the user\n"
-	         "%s"
-	         "\n"),
+	FILE *usageout = (E_SUCCESS != status) ? stderr : stdout;
+	(void) fprintf (usageout,
+	                _("Usage: %s [options] LOGIN\n"
+	                  "       %s -D\n"
+	                  "       %s -D [options]\n"
+	                  "\n"
+	                  "Options:\n"),
+	                Prog, Prog, Prog);
+	(void) fputs (_("  -b, --base-dir BASE_DIR       base directory for the home directory of the\n"
+	                "                                new account\n"), usageout);
+	(void) fputs (_("  -c, --comment COMMENT         GECOS field of the new account\n"), usageout);
+	(void) fputs (_("  -d, --home-dir HOME_DIR       home directory of the new account\n"), usageout);
+	(void) fputs (_("  -D, --defaults                print or change default useradd configuration\n"), usageout);
+	(void) fputs (_("  -e, --expiredate EXPIRE_DATE  expiration date of the new account\n"), usageout);
+	(void) fputs (_("  -f, --inactive INACTIVE       password inactivity period of the new account\n"), usageout);
+	(void) fputs (_("  -g, --gid GROUP               name or ID of the primary group of the new\n"
+	                "                                account\n"), usageout);
+	(void) fputs (_("  -G, --groups GROUPS           list of supplementary groups of the new\n"
+	                "                                account\n"), usageout);
+	(void) fputs (_("  -h, --help                    display this help message and exit\n"), usageout);
+	(void) fputs (_("  -k, --skel SKEL_DIR           use this alternative skeleton directory\n"), usageout);
+	(void) fputs (_("  -K, --key KEY=VALUE           override /etc/login.defs defaults\n"), usageout);
+	(void) fputs (_("  -l, --no-log-init             do not add the user to the lastlog and\n"
+	                "                                faillog databases\n"), usageout);
+	(void) fputs (_("  -m, --create-home             create the user's home directory\n"), usageout);
+	(void) fputs (_("  -M, --no-create-home          do not create the user's home directory\n"), usageout);
+	(void) fputs (_("  -N, --no-user-group           do not create a group with the same name as\n"
+	                "                                the user\n"), usageout);
+	(void) fputs (_("  -o, --non-unique              allow to create users with duplicate\n"
+	                "                                (non-unique) UID\n"), usageout);
+	(void) fputs (_("  -p, --password PASSWORD       encrypted password of the new account\n"), usageout);
+	(void) fputs (_("  -r, --system                  create a system account\n"), usageout);
+	(void) fputs (_("  -R, --root CHROOT_DIR         directory to chroot into\n"), usageout);
+	(void) fputs (_("  -s, --shell SHELL             login shell of the new account\n"), usageout);
+	(void) fputs (_("  -u, --uid UID                 user ID of the new account\n"), usageout);
+	(void) fputs (_("  -U, --user-group              create a group with the same name as the user\n"), usageout);
 #ifdef WITH_SELINUX
-	         _("  -Z, --selinux-user SEUSER     use a specific SEUSER for the SELinux user mapping\n")
-#else
-	         ""
-#endif
-	         );
-	exit (E_USAGE);
+	(void) fputs (_("  -Z, --selinux-user SEUSER     use a specific SEUSER for the SELinux user mapping\n"), usageout);
+#endif				/* WITH_SELINUX */
+	(void) fputs ("\n", usageout);
+	exit (status);
 }
 
 /*
@@ -784,11 +792,11 @@ static void new_spent (struct spwd *spent)
 		spent->sp_inact = scale_age (def_inactive);
 		spent->sp_expire = scale_age (user_expire);
 	} else {
-		spent->sp_min = scale_age (-1);
-		spent->sp_max = scale_age (-1);
-		spent->sp_warn = scale_age (-1);
-		spent->sp_inact = scale_age (-1);
-		spent->sp_expire = scale_age (-1);
+		spent->sp_min = -1;
+		spent->sp_max = -1;
+		spent->sp_warn = -1;
+		spent->sp_inact = -1;
+		spent->sp_expire = -1;
 	}
 	spent->sp_flag = SHADOW_SP_FLAG_UNSET;
 }
@@ -817,6 +825,8 @@ static void grp_update (void)
 	/*
 	 * Scan through the entire group file looking for the groups that
 	 * the user is a member of.
+	 * FIXME: we currently do not check that all groups of user_groups
+	 *        were completed with the new user.
 	 */
 	for (gr_rewind (), grp = gr_next (); NULL != grp; grp = gr_next ()) {
 
@@ -890,6 +900,10 @@ static void grp_update (void)
 		/*
 		 * See if the user specified this group as one of their
 		 * concurrent groups.
+		 * FIXME: is it really needed?
+		 *        This would be important only if the group is in
+		 *        user_groups. All these groups should be checked
+		 *        for existence with gr_locate already.
 		 */
 		if (gr_locate (sgrp->sg_name) == NULL) {
 			continue;
@@ -968,38 +982,39 @@ static void process_flags (int argc, char **argv)
 		 */
 		int c;
 		static struct option long_options[] = {
-			{"base-dir", required_argument, NULL, 'b'},
-			{"comment", required_argument, NULL, 'c'},
-			{"home-dir", required_argument, NULL, 'd'},
-			{"defaults", no_argument, NULL, 'D'},
-			{"expiredate", required_argument, NULL, 'e'},
-			{"inactive", required_argument, NULL, 'f'},
-			{"gid", required_argument, NULL, 'g'},
-			{"groups", required_argument, NULL, 'G'},
-			{"help", no_argument, NULL, 'h'},
-			{"skel", required_argument, NULL, 'k'},
-			{"key", required_argument, NULL, 'K'},
-			{"create-home", no_argument, NULL, 'm'},
-			{"no-create-home", no_argument, NULL, 'M'},
-			{"no-log-init", no_argument, NULL, 'l'},
-			{"no-user-group", no_argument, NULL, 'N'},
-			{"non-unique", no_argument, NULL, 'o'},
-			{"password", required_argument, NULL, 'p'},
-			{"system", no_argument, NULL, 'r'},
-			{"shell", required_argument, NULL, 's'},
+			{"base-dir",       required_argument, NULL, 'b'},
+			{"comment",        required_argument, NULL, 'c'},
+			{"home-dir",       required_argument, NULL, 'd'},
+			{"defaults",       no_argument,       NULL, 'D'},
+			{"expiredate",     required_argument, NULL, 'e'},
+			{"inactive",       required_argument, NULL, 'f'},
+			{"gid",            required_argument, NULL, 'g'},
+			{"groups",         required_argument, NULL, 'G'},
+			{"help",           no_argument,       NULL, 'h'},
+			{"skel",           required_argument, NULL, 'k'},
+			{"key",            required_argument, NULL, 'K'},
+			{"no-log-init",    no_argument,       NULL, 'l'},
+			{"create-home",    no_argument,       NULL, 'm'},
+			{"no-create-home", no_argument,       NULL, 'M'},
+			{"no-user-group",  no_argument,       NULL, 'N'},
+			{"non-unique",     no_argument,       NULL, 'o'},
+			{"password",       required_argument, NULL, 'p'},
+			{"system",         no_argument,       NULL, 'r'},
+			{"root",           required_argument, NULL, 'R'},
+			{"shell",          required_argument, NULL, 's'},
+			{"uid",            required_argument, NULL, 'u'},
+			{"user-group",     no_argument,       NULL, 'U'},
 #ifdef WITH_SELINUX
-			{"selinux-user", required_argument, NULL, 'Z'},
-#endif
-			{"uid", required_argument, NULL, 'u'},
-			{"user-group", no_argument, NULL, 'U'},
+			{"selinux-user",   required_argument, NULL, 'Z'},
+#endif				/* WITH_SELINUX */
 			{NULL, 0, NULL, '\0'}
 		};
 		while ((c = getopt_long (argc, argv,
 #ifdef WITH_SELINUX
-		                         "b:c:d:De:f:g:G:k:K:lmMNop:rs:u:UZ:",
-#else
-		                         "b:c:d:De:f:g:G:k:K:lmMNop:rs:u:U",
-#endif
+		                         "b:c:d:De:f:g:G:hk:K:lmMNop:rR:s:u:UZ:",
+#else				/* !WITH_SELINUX */
+		                         "b:c:d:De:f:g:G:hk:K:lmMNop:rR:s:u:U",
+#endif				/* !WITH_SELINUX */
 		                         long_options, NULL)) != -1) {
 			switch (c) {
 			case 'b':
@@ -1036,14 +1051,14 @@ static void process_flags (int argc, char **argv)
 				break;
 			case 'D':
 				if (anyflag) {
-					usage ();
+					usage (E_USAGE);
 				}
 				Dflg = true;
 				break;
 			case 'e':
 				if ('\0' != *optarg) {
 					user_expire = strtoday (optarg);
-					if (user_expire == -1) {
+					if (user_expire < -1) {
 						fprintf (stderr,
 						         _("%s: invalid date '%s'\n"),
 						         Prog, optarg);
@@ -1054,9 +1069,10 @@ static void process_flags (int argc, char **argv)
 				}
 
 				/*
-				 * -e "" is allowed - it's a no-op without /etc/shadow
+				 * -e "" is allowed without /etc/shadow
+				 * (it's a no-op in such case)
 				 */
-				if (('\0' != *optarg) && !is_shadow_pwd) {
+				if ((-1 != user_expire) && !is_shadow_pwd) {
 					fprintf (stderr,
 					         _("%s: shadow passwords required for -e\n"),
 					         Prog);
@@ -1073,7 +1089,7 @@ static void process_flags (int argc, char **argv)
 					fprintf (stderr,
 					         _("%s: invalid numeric argument '%s'\n"),
 					         Prog, optarg);
-					usage ();
+					exit (E_BAD_ARG);
 				}
 				/*
 				 * -f -1 is allowed
@@ -1113,7 +1129,7 @@ static void process_flags (int argc, char **argv)
 				Gflg = true;
 				break;
 			case 'h':
-				usage ();
+				usage (E_SUCCESS);
 				break;
 			case 'k':
 				def_template = optarg;
@@ -1166,6 +1182,8 @@ static void process_flags (int argc, char **argv)
 			case 'r':
 				rflg = true;
 				break;
+			case 'R': /* no-op, handled in process_root_flag () */
+				break;
 			case 's':
 				if (   ( !VALID (optarg) )
 				    || (   ('\0' != optarg[0])
@@ -1197,7 +1215,6 @@ static void process_flags (int argc, char **argv)
 			case 'Z':
 				if (is_selinux_enabled () > 0) {
 					user_selinux = optarg;
-					Zflg = true;
 				} else {
 					fprintf (stderr,
 					         _("%s: -Z requires SELinux enabled kernel\n"),
@@ -1206,9 +1223,9 @@ static void process_flags (int argc, char **argv)
 					exit (E_BAD_ARG);
 				}
 				break;
-#endif
+#endif				/* WITH_SELINUX */
 			default:
-				usage ();
+				usage (E_USAGE);
 			}
 			anyflag = true;
 		}
@@ -1227,31 +1244,31 @@ static void process_flags (int argc, char **argv)
 		fprintf (stderr,
 		         _("%s: %s flag is only allowed with the %s flag\n"),
 		         Prog, "-o", "-u");
-		usage ();
+		usage (E_USAGE);
 	}
 	if (kflg && !mflg) {
 		fprintf (stderr,
 		         _("%s: %s flag is only allowed with the %s flag\n"),
 		         Prog, "-k", "-m");
-		usage ();
+		usage (E_USAGE);
 	}
 	if (Uflg && gflg) {
 		fprintf (stderr,
 		         _("%s: options %s and %s conflict\n"),
 		         Prog, "-U", "-g");
-		usage ();
+		usage (E_USAGE);
 	}
 	if (Uflg && Nflg) {
 		fprintf (stderr,
 		         _("%s: options %s and %s conflict\n"),
 		         Prog, "-U", "-N");
-		usage ();
+		usage (E_USAGE);
 	}
 	if (mflg && Mflg) {
 		fprintf (stderr,
 		         _("%s: options %s and %s conflict\n"),
 		         Prog, "-m", "-M");
-		usage ();
+		usage (E_USAGE);
 	}
 
 	/*
@@ -1260,15 +1277,15 @@ static void process_flags (int argc, char **argv)
 	 */
 	if (Dflg) {
 		if (optind != argc) {
-			usage ();
+			usage (E_USAGE);
 		}
 
-		if (uflg || oflg || Gflg || dflg || cflg || mflg) {
-			usage ();
+		if (uflg || Gflg || dflg || cflg || mflg) {
+			usage (E_USAGE);
 		}
 	} else {
 		if (optind != argc - 1) {
-			usage ();
+			usage (E_USAGE);
 		}
 
 		user_name = argv[optind];
@@ -1314,7 +1331,7 @@ static void process_flags (int argc, char **argv)
 	if (!rflg) {
 		/* for system accounts defaults are ignored and we
 		 * do not create a home dir */
-		if (getdef_bool("CREATE_HOME")) {
+		if (getdef_bool ("CREATE_HOME")) {
 			mflg = true;
 		}
 	}
@@ -1435,21 +1452,8 @@ static void open_files (void)
 		fprintf (stderr, _("%s: cannot open %s\n"), Prog, pw_dbname ());
 		fail_exit (E_PW_UPDATE);
 	}
-	if (is_shadow_pwd) {
-		if (spw_lock () == 0) {
-			fprintf (stderr,
-			         _("%s: cannot lock %s; try again later.\n"),
-			         Prog, spw_dbname ());
-			fail_exit (E_PW_UPDATE);
-		}
-		spw_locked = true;
-		if (spw_open (O_RDWR) == 0) {
-			fprintf (stderr,
-			         _("%s: cannot open %s\n"),
-			         Prog, spw_dbname ());
-			fail_exit (E_PW_UPDATE);
-		}
-	}
+
+	/* shadow file will be opened by open_shadow(); */
 
 	/*
 	 * Lock and open the group file.
@@ -1484,6 +1488,26 @@ static void open_files (void)
 #endif
 }
 
+static void open_shadow (void)
+{
+	if (!is_shadow_pwd) {
+		return;
+	}
+	if (spw_lock () == 0) {
+		fprintf (stderr,
+		         _("%s: cannot lock %s; try again later.\n"),
+		         Prog, spw_dbname ());
+		fail_exit (E_PW_UPDATE);
+	}
+	spw_locked = true;
+	if (spw_open (O_RDWR) == 0) {
+		fprintf (stderr,
+		         _("%s: cannot open %s\n"),
+		         Prog, spw_dbname ());
+		fail_exit (E_PW_UPDATE);
+	}
+}
+
 static char *empty_list = NULL;
 
 /*
@@ -1497,7 +1521,14 @@ static void new_grent (struct group *grent)
 {
 	memzero (grent, sizeof *grent);
 	grent->gr_name = (char *) user_name;
-	grent->gr_passwd = SHADOW_PASSWD_STRING;	/* XXX warning: const */
+#ifdef  SHADOWGRP
+	if (is_shadow_grp) {
+		grent->gr_passwd = SHADOW_PASSWD_STRING;	/* XXX warning: const */
+	} else
+#endif				/* SHADOWGRP */
+	{
+		grent->gr_passwd = "!";	/* XXX warning: const */
+	}
 	grent->gr_gid = user_gid;
 	grent->gr_mem = &empty_list;
 }
@@ -1716,32 +1747,6 @@ static void usr_update (void)
 	}
 }
 
-#ifdef WITH_SELINUX
-static void selinux_update_mapping (void) {
-	if (is_selinux_enabled () <= 0) return;
-
-	if (*user_selinux) { /* must be done after passwd write() */
-		const char *argv[7];
-		argv[0] = "/usr/sbin/semanage";
-		argv[1] = "login";
-		argv[2] = "-a";
-		argv[3] = "-s";
-		argv[4] = user_selinux;
-		argv[5] = user_name;
-		argv[6] = NULL;
-		if (safe_system (argv[0], argv, NULL, 0)) {
-			fprintf (stderr,
-			         _("%s: warning: the user name %s to %s SELinux user mapping failed.\n"),
-			         Prog, user_name, user_selinux);
-#ifdef WITH_AUDIT
-			audit_logger (AUDIT_ADD_USER, Prog,
-			              "adding SELinux user mapping",
-			              user_name, (unsigned int) user_id, 0);
-#endif
-		}
-	}
-}
-#endif
 /*
  * create_home - create the user's home directory
  *
@@ -1753,7 +1758,9 @@ static void create_home (void)
 {
 	if (access (user_home, F_OK) != 0) {
 #ifdef WITH_SELINUX
-		selinux_file_context (user_home);
+		if (set_selinux_file_context (user_home) != 0) {
+			fail_exit (E_HOMEDIR);
+		}
 #endif
 		/* XXX - create missing parent directories.  --marekm */
 		if (mkdir (user_home, 0) != 0) {
@@ -1780,7 +1787,9 @@ static void create_home (void)
 #endif
 #ifdef WITH_SELINUX
 		/* Reset SELinux to create files with default contexts */
-		setfscreatecon (NULL);
+		if (reset_selinux_file_context () != 0) {
+			fail_exit (E_HOMEDIR);
+		}
 #endif
 	}
 }
@@ -1794,13 +1803,14 @@ static void create_home (void)
  */
 static void create_mail (void)
 {
-	char *spool, *file;
-	int fd;
-	struct group *gr;
-	gid_t gid;
-	mode_t mode;
-
 	if (strcasecmp (create_mail_spool, "yes") == 0) {
+		const char *spool;
+		char *file;
+		int fd;
+		struct group *gr;
+		gid_t gid;
+		mode_t mode;
+
 		spool = getdef_str ("MAIL_DIR");
 		if (NULL == spool) {
 			spool = "/var/mail";
@@ -1846,10 +1856,6 @@ int main (int argc, char **argv)
 #endif				/* USE_PAM */
 #endif				/* ACCT_TOOLS_SETUID */
 
-#ifdef WITH_AUDIT
-	audit_help_open ();
-#endif
-
 	/*
 	 * Get my name so that I can use it to report errors.
 	 */
@@ -1859,7 +1865,12 @@ int main (int argc, char **argv)
 	(void) bindtextdomain (PACKAGE, LOCALEDIR);
 	(void) textdomain (PACKAGE);
 
+	process_root_flag ("-R", argc, argv);
+
 	OPENLOG ("useradd");
+#ifdef WITH_AUDIT
+	audit_help_open ();
+#endif
 
 	sys_ngroups = sysconf (_SC_NGROUPS_MAX);
 	user_groups = (char **) xmalloc ((1 + sys_ngroups) * sizeof (char *));
@@ -1901,13 +1912,16 @@ int main (int argc, char **argv)
 		retval = pam_acct_mgmt (pamh, 0);
 	}
 
-	if (NULL != pamh) {
-		(void) pam_end (pamh, retval);
-	}
 	if (PAM_SUCCESS != retval) {
-		fprintf (stderr, _("%s: PAM authentication failed\n"), Prog);
+		fprintf (stderr, _("%s: PAM: %s\n"),
+		         Prog, pam_strerror (pamh, retval));
+		SYSLOG((LOG_ERR, "%s", pam_strerror (pamh, retval)));
+		if (NULL != pamh) {
+			(void) pam_end (pamh, retval);
+		}
 		fail_exit (1);
 	}
+	(void) pam_end (pamh, retval);
 #endif				/* USE_PAM */
 #endif				/* ACCT_TOOLS_SETUID */
 
@@ -1996,6 +2010,18 @@ int main (int argc, char **argv)
 		}
 	}
 
+#ifdef WITH_TCB
+	if (getdef_bool ("USE_TCB")) {
+		if (shadowtcb_create (user_name, user_id) == SHADOWTCB_FAILURE) {
+			fprintf (stderr,
+			         _("%s: Failed to create tcb directory for %s\n"),
+			         Prog, user_name);
+			fail_exit (E_UID_IN_USE);
+		}
+	}
+#endif
+	open_shadow ();
+
 	/* do we have to add a group for that user? This is why we need to
 	 * open the group files in the open_files() function  --gafton */
 	if (Uflg) {
@@ -2013,7 +2039,8 @@ int main (int argc, char **argv)
 	if (mflg) {
 		create_home ();
 		if (home_added) {
-			copy_tree (def_template, user_home, user_id, user_gid);
+			copy_tree (def_template, user_home, false, false,
+			           (uid_t)-1, user_id, (gid_t)-1, user_gid);
 		} else {
 			fprintf (stderr,
 			         _("%s: warning: the home directory already exists.\n"
@@ -2024,15 +2051,27 @@ int main (int argc, char **argv)
 	}
 
 	/* Do not create mail directory for system accounts */
-	if( !rflg ) {
+	if (!rflg) {
 		create_mail ();
 	}
 
 	close_files ();
 
 #ifdef WITH_SELINUX
-	selinux_update_mapping ();
-#endif
+	if (Zflg) {
+		if (set_seuser (user_name, user_selinux) != 0) {
+			fprintf (stderr,
+			         _("%s: warning: the user name %s to %s SELinux user mapping failed.\n"),
+			         Prog, user_name, user_selinux);
+#ifdef WITH_AUDIT
+			audit_logger (AUDIT_ADD_USER, Prog,
+			              "adding SELinux user mapping",
+			              user_name, (unsigned int) user_id, 0);
+#endif				/* WITH_AUDIT */
+			fail_exit (E_SE_UPDATE);
+		}
+	}
+#endif				/* WITH_SELINUX */
 
 	nscd_flush_cache ("passwd");
 	nscd_flush_cache ("group");
